function neweeg = set_inject_events(eeg,varargin)
% Result-Set = set_inject_events(Input-Set, Options...)
% Inject runs of events in specific segments of a continuous dataset.
%
% Almost all real-time inference in BCIs is done on the basis of epochs, and epochs are most conveniently created relative to 
% certain ("time-locking") events in the data. This function allows to cover periods of continuous data with events, at regular or random intervals,
% so that epochs covering these ranges can subsequently be extracted. What periods shall be populated with events can be flexibly specified.
%
% In:
%   Input-Set   : continuous EEG dataset
%
%   Options...  : name-value pairs controlling the injection
%                 'segment'   : segment specification. cell array of one of the following forms (lats in seconds):
%                               note: the ordering of time values w.r.t. eventtype values w.r.t. cellarray values in the subsequent 
%                                     specifications is arbitrary, whereas the ordering of time values w.r.t. each other is relevant 
%                                     (first time value shall be lower than second time value); the ordering of eventtypes w.r.t. each 
%                                     other is also relevant.
%                               * {absolute_time absolute_time}:
%                                 segment specified using two time points
%                               * {event_type relative_time relative_time} / {relative_time  event_type relative_time} / 
%                                 {relative_time relative_time event_type}:
%                                 here, the segment is relative to some event (of a given type), in between the time interval given by the 
%                                 first and second time value
%                               * {event_type relative_time relative_time event_type} / {relative_time event_type event_type relative_time} / ...
%                                 here, the segment is in between two immediately successive events of the given types (any intervals with other 
%                                 events in between the specified ones are not considered for injection), constrained by the relative lats 
%                                 for each event
%                               * {event_type relative_time {ignore_type1,ignore_type2,...} relative_time event_type}
%                                 as above, except that intermediate events of type ignore_type1/ignore_type2/etc. are ignored
%                                 (if the in-between cell array is empty, any other events are ignored)
%
%                 'event'     : the inserted event type string, or alternatively a template event struct 
%                                (default: "mytype_i", when injected relative to an event of type "mytype" or
%                                 "mytype1_mytype2", when injected in between two events of type "mytype1" and "mytype2")
%                               * if a string is specified, all event fields besides the 'type', 'latency' and 'duration' fields 
%                                  contained in the data will be left empty 
%                               * if a struct is specified, the 'latency' field will be substituted appropriately
%
%                 'count'     : number of events inserted (default: 1)
%
%                 'counting'  : what count means for any given segment, either 'total' or 'persecond' (default: 'total')
%
%                 'placement' : how the injected events should be placed, either 'random' or 'equidistant' (default: 'random')
%
%                 'repeatable': whether the randomization procedure shall give repeatable results (default: 1); different numbers (aside from 0)
%                               give different repeatable runs, i.e. the value determines the randseed
%
%                 'limits'    : optional time limits (in seconds) to constrain event placement
%
% Out:
%   Output-Set  : continuous EEG dataset with new events injected
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-05-24
global bcilab;

if ~exp_begindef('editing') return; end

if isfield(eeg,'epoch') && ~isempty(eeg.epoch)
    error('the dataset appears to contain epochs: only continuous dataset are supported for now.'); end

% get options
opts = hlp_varargin2struct(varargin,'segment','<mandatory>', 'event',[], 'count',1, 'counting','total', 'placement','random', 'repeatable',1, ...
    'srate',eeg.srate, 'limits',[eeg.xmin,eeg.xmax]);
opts.limits = opts.limits*opts.srate;

% parse the segment specification
mrks = {};
lats = [];
ignored = [];
for i=1:length(opts.segment)
    if ischar(opts.segment{i})
        mrks{end+1} = opts.segment{i};
    elseif iscell(opts.segment{i})
        ignored = [ignored opts.segment{i}];
    else
        lats(end+1) = opts.segment{i};
    end
end
lats = lats*eeg.srate;

% init randomization
if strcmp(opts.placement,'random') && opts.repeatable
    if hlp_version < 707
        % save & override RNG state
        randstate = rand('state'); %#ok<*RAND>
        rand('state',5182+opts.repeatable); %#ok<RAND>
    else
        % create a legacy-compatible RandStream
        bcilab.temp.randstream_inject_events = RandStream('swb2712','Seed',5182+opts.repeatable);
    end
end

neweeg = eeg;
if isempty(mrks)
    % --- inject using absolute lats ---
    if isempty(opts.event)
        error('an event type must be specified'); end
    if ischar(opts.event)
        opts.event = make_default_event(eeg.event,opts.event); end
    neweeg = perform_injection(neweeg,lats,opts);
elseif length(mrks) == 1
    % --- inject relative to one single marker ---
    if isempty(opts.event)
        opts.event = [mrks{1} '_i']; end
    if ischar(opts.event)
        opts.event = make_default_event(eeg.event,opts.event); end
    for e=1:length(eeg.event)
        if strcmp(eeg.event(e).type,mrks{1})
            neweeg = perform_injection(neweeg,eeg.event(e).latency+lats,opts); end
    end
elseif length(mrks) == 2
    % --- inject in between two successive mrks ---
    if isempty(opts.event)
        opts.event = [mrks{1} '_' mrks{2}]; end
    if ischar(opts.event)
        opts.event = make_default_event(eeg.event,opts.event); end
    startlat = [];
    for e=1:length(eeg.event)
        % ignorance turned off but intermediate marker found?
        if ~iscell(ignored) && ~strcmp(eeg.event(e).type,mrks{2})
            startlat = []; end
        % selective ignorance turned on but non-ignored intermediate marker found?
        if iscell(ignored) && ~isempty(ignored) && ~any(strcmp(eeg.event(e).type,[ignored mrks(2)]))
            startlat = []; end
        % found the begin-marker: open a new segment
        if strcmp(eeg.event(e).type,mrks{1})
            startlat = eeg.event(e).latency; end
        % reached the end-marker & have an open segment: inject.
        if ~isempty(startlat) && strcmp(eeg.event(e).type,mrks{2})
            neweeg = perform_injection(neweeg,[startlat+lats(1),eeg.event(e).latency+lats(2)],opts); end
    end
end
% sort the events by latency...
neweeg.event = neweeg.event(hlp_getresult(2,@sort,[neweeg.event.latency]));
% reproduce urevents if necessary
if isempty(neweeg.urevent)
    neweeg = eeg_checkset(neweeg,'makeur'); end


% conclude randomization
if strcmp(opts.placement,'random') && opts.repeatable && hlp_version < 707
    % restore saved RNG state
    rand('state',randstate); %#ok<RAND>
end

exp_enddef;

function [eeg,coverage] = perform_injection(eeg,ival,opts)
global bcilab;
coverage = 0;
% sanity check
if length(ival) == 2 && ival(1) <= ival(2)
    coverage = ival(2)-ival(1);
    % hande the spacing method
    if strcmp(opts.counting,'persecond')
        opts.count = max(1,round(opts.count*(ival(2)-ival(1))/opts.srate)); end
    % handle the placement method
    if strcmp(opts.placement,'equidistant')
        stepsize = (ival(2)-ival(1))/(opts.count-1);
        ival = round(ival(1):stepsize:ival(2));
    elseif strcmp(opts.placement,'random')
        if hlp_version < 707
            positions = rand(1,opts.count);
        else
            positions = rand(bcilab.temp.randstream_inject_events,1,opts.count);
        end
        ival = round(positions*(ival(2)-ival(1))+ival(1));
    else
        error('unsupported placement scheme specified');
    end
    % compute the individual latencies
    lats = ival(ival>=opts.limits(1) & ival<=opts.limits(2));
    if ~isempty(eeg.urevent)
        eeg.urevent = []; end
    if isempty(eeg.event)
        eeg.event = setfield(setfield(opts.event,'latency',1),'type','dummy'); end; %#ok<*SFLD>
    for l = lats
        eeg.event(end+1) = setfield(opts.event,'latency',l); end %#ok<SFLD>
end

% create a default event from an event array
function evt = make_default_event(evts,type)
if isempty(evts)
    evt = struct('type',{type},'latency',{[]},'duration',{1},'urevent',{[]});
else
    evt = evts(1);
    for fn=fieldnames(evt)'
        evt.(fn{1}) = []; end
    evt.type = type;
    if isfield(evt,'duration')
        evt.duration = 1; end
end
