function eeg = set_makepos(eeg,timewnd,eventtypes)
% Out-EEG = set_makepos(Input-EEG,Time-Window,Event-Types)
% Create epochs relative to specified markers and assign the epochs' target variable.
%
% This function is a convenience extension to EEGLAB's pop_epoch, which implicitly assigns a target value to each epoch. It is the default
% mechanism by which paradigms convert continuous data to epoched (and target-annotated) data. By default, the order of specified event types 
% determines to what target values the respective events map. Event types can be listed in groups (cell arrays), so that multiple event types can  
% map to the same target variable.
%
% In:
%   Input-EEG       :   continuous dataset from which epochs shall be extracted
%
%   Time-Window     :   time window relative to the events (forming the extracted epoch), in seconds
%
%   Event-Types     :   Types of the events/markers for which epochs shall be extracted; cell array of strings.
%
%                       The order in which marker types are supplied can be utilitzed conveniently to have a target
%                       value assigned to each created epoch. For example, if this is {'xxx' 'dfg' 'yyy'}, all epochs 
%                       with time-locking event type 'yyy' will be assigned a target value of 3, while those with event-type 'dfg' 
%                       will receive target value of 2, etc. (i.e., the target value is the index of the time-locking event type, 
%                       as it appears in the the Event-Types cell array).
%
%                       To assign the same target value to events of multiple types, group them into a nested cell array, e.g.,
%                       {'xxx' {'dfg' 'blah'} 'yyy'} -- here, the assignment from time-locking event type to target value is as follows:
%                        xxx -> 1, dfg -> 2, blah -> 2, yyy -> 3.
%
% Out:
%   Out-EEG  :  epoched dataset, with target variable assigned; changes:
%               - data is reformatted into a 3d array
%               - xmin & xmax indicate the epoch bounds in seconds
%               - events are rewritten to be relative to epochs (urevents unaffected)
%               - times contains the time points inside the epoch
%
% Examples:
%   epos = set_makepos(myset,[1 3],{'left-motor-imagery','right-motor-imagery'});
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-01

if ~exp_begindef('filter') return; end
epolength = @(wnd,srate) round(wnd(2)*srate) - round(wnd(1)*srate);

if epolength(timewnd,eeg.srate) >= size(eeg.data,2) && isfield(eeg,'smin')
    % --- online case: 
    
    % fix up xmin and xmax for the current epoch
    eeg.xmin = timewnd(1);
    eeg.xmax = timewnd(1) + (size(eeg.data,2)-1)/eeg.srate;    
    % add an meaningful epoch field; this contains at least the latency
    eeg.epoch = struct('latency', {eeg.smin + eeg.xmin*eeg.srate});
else
    % --- regular (offline) case: extract epochs...
    disp('creating epochs...');
    
    % create a flattened eventtype list and a target value list,
    % so we can compute the proper target value per epoch
    if any(cellfun(@iscell,eventtypes))
        targetlist = [];
        typelist = {};
        for i=1:length(eventtypes)
            if ischar(eventtypes{i})
                targetlist(end+1) = i;
                typelist{end+1} = eventtypes{i};
            elseif iscell(eventtypes{i})
                targetlist(end+1:end+length(eventtypes{i})) = i;
                typelist(end+1:end+length(eventtypes{i})) = eventtypes{i};
            else
                error('unsupported event type syntax');
            end
        end           
    else
        typelist = eventtypes;
        targetlist = 1:length(eventtypes);
    end
    
    % sort events (as expected by pop_epoch)
    eeg.event = eeg.event(hlp_getresult(2,@sort,[eeg.event.latency]));

    % retain original event types and identify matching event indices
    alltypes = {eeg.event.type};
    alllatencies = [eeg.event.latency];
    matchidx = [];
    for t=1:length(typelist)
        matchidx(strcmp(typelist{t},alltypes)) = 1; end
    matchidx = find(matchidx);
    
    % extract epochs
    [eeg, evtindices] = pop_epoch(eeg, typelist, timewnd);

    % make sure that we can compute the length of the epochs in points
    if epolength(timewnd,eeg.srate) ~= size(eeg.data,2)
        warning('BCILAB:set_makepos:cannot_compute_epolen',['Could not compute the appropriate epoch length; ' ...
        'online operation will not work. (this should never happen)']); end
    
    % assign target values and epoch latencies    
    for i=1:length(evtindices)
        eeg.epoch(i).target = targetlist(strcmp(alltypes{matchidx(evtindices(i))},typelist));
        eeg.epoch(i).latency = alllatencies(matchidx(evtindices(i)));
    end
end

exp_enddef;
