function eeg = set_new(varargin)
% Dataset = set_new(Arguments...)
% Create a new EEGLAB data set from individual fields.
%
% In:
%   Fields  : Pairs of field names and field values to add to the data set. fields not specified are taken from eeg_emptyset, 
%             later fields override earlier fields; giving a struct in place of a name-value pair is equivalent to writing 
%             out all the struct fieldnames and respective values. fields that can be derived from others are derived.
%
%   `         optional special semantics:
%             * 'chanlocs' can be specified as cell-string array, and is generally completed using a default lookup
%             * 'data' can be specified as a cell array of data arrays, then concatenated across epochs,
%                      and with .epoch.target derived as the index of the cell which contained the epoch in question.
%             * 'bcilab.online_expression' can be specified to override the online processing description
%
% Out:
%   Dataset : newly created EEGLAB set
% 
% Example:
%   myset = set_new('data',randn(3,100000), 'srate',1000,'chanlocs',struct('labels',{'A','B','C'}));
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-05-28

if ~exp_begindef('filter') return; end

% construct the dataset from arguments and defaults
eeg = hlp_varargin2struct(varargin, eeg_emptyset, 'setname','new set', 'bcilab.online_expression',@rawdata);

% rewrite cell array data
if iscell(eeg.data)
    data = eeg.data;
    eeg.data = [];
    for c=1:length(data)
        if isnumeric(data{c}) && ~isempty(data{c})
            eeg.data = cat(3, eeg.data, data{c});
            for i=size(eeg.data,3)-size(data{c},3)+1:size(eeg.data,3)
                eeg.epoch(i).target = c; end
        end
    end    
end

% rewrite cell array chanlocs
if iscell(eeg.chanlocs)
    eeg.chanlocs = struct('labels',eeg.chanlocs); end;

% derive additional chanlocs fields
if ~isfield(eeg.chanlocs,'X') || ~isvector([eeg.chanlocs.X])
    try eeg.chanlocs = hlp_microcache('setnew',@pop_chanedit,eeg.chanlocs,'lookup','standard-10-5-cap385.elp'); catch, end
end

% derive xmax, nbchan, pnts, trials
[eeg.nbchan,eeg.pnts, eeg.trials] = size(eeg.data);
eeg.xmax = eeg.xmin + (eeg.pnts-1)/eeg.srate;

% derive additional event & urevent info
if isfield(eeg,'event')
    eeg = eeg_checkset(eeg,'eventconsistency'); 
    eeg = eeg_checkset(eeg,'makeur'); 
end

% add epoch.latency if possible
if ~isfield(eeg.epoch,'latency')
    for i=1:length(eeg.epoch)
        tle = [eeg.epoch(i).eventlatency{:}]==0;
        if any(tle)
            eeg.epoch(i).latency = b.event(b.epoch(i).event(tle)).latency; end
    end
end

% do minimal consistency checks
if ~isempty(eeg.chanlocs) && ~isempty(eeg.data) && (length(eeg.chanlocs) ~= eeg.nbchan)
    error('The number of supplied channel locations does not match the number of channels in the data.'); end
if isfield(eeg,'epoch') && ~isempty(eeg.epoch) && length(eeg.epoch) ~= size(eeg.data,3)
    error('The number of data epochs does not match the number of entries in the epoch field'); end

% add online expression
if isfield(eeg,'bcilab') && isfield(eeg.bcilab,'online_expression')
    exp_enddef('add_online',eeg.bcilab.online_expression);
else
    exp_enddef('add_online',@rawdata);
end
