function eeg = set_picktrials(eeg, varargin)
% Result-Set = set_picktrials(Input-Set, Arguments...)
% Select trials whose target variable satisfies a certain condition.
%
% This function can be used to obtain a subset of the dataset, reduced to those epochs that have target values matching the supplied criteria.
%
% In:
%   Input-Set   :   Epoched dataset with a target variable.
%
%   Arguments   :   selection criteria (name-value pairs):
%                    'value': target value or cell array of alternative target values 
%                             -> select by value 
%                    'range': 2-element cell array of the lower and higher end of the range
%                             -> select by range
%                    'rank' : value rank or cell array of alternative values ranks
%                             -> select by rank
%                    currently, only one criterion is supported at a time
%
% Out:
%   Result-Set  :  Epoched dataset which contains only the selected trials.
%
% Notes:
%   This function (in rank-selection mode) is used by methods that need to extract the trials of the first class, second class, etc. from the data.
% 
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-01

% sort name-value pairs according to name
varargin = hlp_sort_namevalues(varargin{:});
if ~exp_begindef('editing') return; end

opts = hlp_varargin2struct(varargin, 'value',[], 'range',[], 'rank',[]);

if ~isfield(eeg,'epoch')
    error('the dataset should contain epochs'); end
if ~isfield(eeg.epoch,'target')
    error('the dataset does not contain an epoch field named target (denoting the target variable)'); end
if ~isempty(opts.rank) + ~isempty(opts.value) + ~isempty(opts.range) ~= 1
    error('exactly a single criterion is allowed at a time'); end

variable = vertcat(eeg.epoch.target);

if ~isempty(opts.rank)
    % do rank-based selection
    if size(variable,2) > 1
        error('rank-based selection is only supported for one-dimensional target values'); end
    % rank-transform the variable
    [a,b,variable] = unique(variable); %#ok<ASGLU>
    opts.value = opts.rank;
end

if ~isempty(opts.value)
    % do value-based selection
    if ~iscell(opts.value)
        opts.value = {opts.value}; end
    inds = [];
    for i=1:length(opts.value)
        inds = [inds; find(prod(single(variable == repmat(opts.value{i},size(variable,1),1)),2))]; end        
end

if ~isempty(opts.range)
    % do range-based selection
    if ~iscell(opts.range)
        error('range based testing requires pairs (two-element cell arrays) of values'); end
    if ~iscell(opts.range{1})
        opts.range = {opts.range}; end
    inds = [];
    for i=1:length(opts.range)
        if length(opts.range{i}) ~= 2
            error('range based testing requires pairs (two-element cell arrays) of values'); end
        inds = [inds; intersect(find(prod(single(variable >= repmat(opts.range{i}{1},size(variable,1),1)),2)), ...
                                find(prod(single(variable <= repmat(opts.range{i}{2},size(variable,1),1)),2)))]; 
    end
end

%eeg = pop_select(eeg,'trial',inds');
evalc('eeg = pop_select(eeg,''trial'',inds'');');

exp_enddef;
