function immediate = exp_begindef(setting, varargin)
% Immediate-Eval = exp_begindef(Setting, Attributes...)
% Begin the definition of an expression-aware function (= a function that can take and/or return expressions).
%
% exp_begindef is part of the lower-level infrastructure of BCILAB, and mainly serves to simplify the contract of Signal Processing filters, and 
% dataset editing functions. It is practically always called according to a close-to-trivial contract, which looks like the following:
%
%   function result = my_function(arg1,arg2,arg3, ...)
%   if ~exp_begindef('type') return; end
%   ... <actual implementation> ...
%   exp_enddef;
% 
% ... where 'type' depends on the type of function being defined. Generally, all functions that operate on datasets and modify the signal content in an
% online-relevant way are of type 'filter', and all functions that operate on datasets in ways which have no meaning during online processing are of type
% 'editing' (which is just a convenience shortcut for writing the function so that it does not modify datasets when called online). See filters/flt_* and
% dataset_ops/set_* for some examples.
%
% For 'filter' functions, the following rules hold: 
% * A filter function should be viewed as the implementation of a node in a filter graph, which receives signal blocks in form of EEGLAB datasets, 
%   and optionally parameters (such as state), and returns a processed version of its input signal block.
% * A filter function will be called offline, on a recorded dataset (passed as a whole), and online, on an arbitrarily short block of data (down to one sample)
% * A filter that is applied to a block and then applied to a subsequent block SHOULD produce blocks that, when concatenated, are identical to what the filter
%   would produce when applied only once to a concatenation of its input blocks (i.e. it should be invariant under block size). 
%   -> To implement this, a filter should generally support continuation: If information from the previous block is necessary to continue on a subsequent block
%      while satisfying invariance, the filter should return that information as a second output, and should be able to accept this type of information
%      as its first non-dataset input. It is then called stateful. A stateful function should not use its first non-dataset input for anything else than state
%      (and it can be expected to be [] when the function call is not a "continued" call).
% * A filter function MUST NOT overwrite/change any of its its non-EEG (and non-State) input arguments, except to achieve behavior described below.
% * A filter function MAY define how its input arguments during an online call (small blocks) shall differ from an offline call (one large block),
%   in one of two ways: 1. override the input arguments during offline processing, to determine their state during online processing
%                       2. pass an 'online' parameter to exp_enddef which contains a custom alternative online expression (which is then invoked online)
% * A filter function MAY be non-invariant under block size if not otherwise possible (e.g. for performance reasons); in this case, it MUST be understood that 
%   the function is doing so. (an example is ICA: adaptive on the first (assumed to be large) block, but not adapting ony any subsequent block)
%
% For 'editing' functions, the following rules hold:
% * An editing function MUST not edit a (streamable) dataset in a way such that subsequent functions will produce different signal values. If it is doing so,
%   it is a 'filter' function. STUDY sets are not considered streamable, only EEG sets are.
%
%
% For the purpose of data processing function declarations, the exp_begindef/exp_enddef contract handles the book-keeping necessary to be able to reproduce 
% the filter online, and offers additional functionality, such as transparent caching of results (for later near-instantaneuous retrieval) and the annotation
% of all datasets with an executable expression which, when invoked, reproduces the dataset 1:1, as well as checks to guarantee that no invalid
% tampering is done by the user on datasets (such as running a forward-backward filter in EEGLAB or manually scaling the data).
%
% In its general form, the functions exp_begindef/exp_enddef are part of the machinery that enables symbolic computation in MATLAB (in the
% style of Mathematica), and the 'filter'/'editing'/'symbolic'/'annotate'/'default' argument is a short-cut which selects default attributes of the function being
% defined (which can be overriden). Some of these attributes are customizable plugin "hooks" (applied to each argument or all inputs or all outputs of the function 
% being defined). 'Filter'/'editing' set such hooks as default attributes, which in turn implement the tamper-proofing checks and some post-processing on the 
% emitted datasets. The post-processing adds fields (subfields of the .bcilab field) to the (otherwise arbitrary) dataset structs that are produced by the 
% function being defined, most importantly the .bcilab.expression, which is an executable expression tree which mirrors each dataset in symbolic form,
% and .bcilab.online_expression, in the case of filters, which defines the arguments to be used for the respective filter during online processing.
% These operations are usually completely transparent to the user (and locked away in the .bcilab field), and make it exceptionally simple to author components
% for the toolbox.
%
% The symbolic-computation functionality implemented by the expressions/exp_* functions is largely non-overlapping with what the MATLAB Symbolic toolbox offers:
% the Symbolic toolbox exclusively implements mathematics (e.g. derivatives, integrals, simplifications), while the expression system of BCILAB exclusively 
% implements expression manipulation, such as substitution, pattern matching, hashing, memoization, and reflection. These facilities allow to implement very 
% high-level behavior in functions such as bci_train, such as generic programming (e.g. utl_crossval, utl_searchmodel, utl_gridsearch being non-intrusively 
% adapted to EEG sets, STUDY sets, etc.), optimizations (e.g. in-memory caching and on-disk caching of intermediate results, computational shortcuts, and more 
% in the future), and more in the future, with relatively little coding effort.
%
% In:
%  Setting        : baseline attribute assignment for the function; must be specified
%                   * defining a function that should not evaluate immediately, but only after evaluation by an exp_eval() clause:
%                     'default' : delayed execution, input fingerprint checks
%                   * defining a function that processes datasets and should have delayed-evaluation semantics:
%                     'filter'  : functions that do non-trivial processing of datasets, and which should support online replication of 
%                                 some offline processing; attributes: delayed (1), add_impure (1), add_online ('endofbody'), 
%                                 also performs EEGLAB-aware input dataset checks & transforms
%                                 note: when an offline processing is replicated online, the function is called successively on 
%                                       short EEGLAB datasets, with inputs identical to the input parameters as they were snapshot
%                                       after execution of the function's body (by exp_enddef)
%                     'editing' : functions that do processing of datasets which has no effect in online processing, i.e. which pass the
%                                 input data through unmodified; attributes: delayed (1), add_impure (1), add_online ('passthrough'), also performs
%                                 EEGLAB-aware input dataset checks
%                   * defining a function that performs purely symbolic manipulations of expressions:
%                     'symbolic': all inputs are held unevaluated, associated (impure) values are ignored (thus no fingerprinting and no memoization);
%                                 attributes: delayed (1), hold ('all'), fingerprints unchecked & passed through, memoize (0)
%                   * defining a function that performs immediate non-evaluation of data structures, and which shall have their results annotated with
%                     an executable (repeatable) description of the processing performed:
%                     'annotate': delayed (0), add_impure (1)
%
%
%   Attributes... : optional name-value pairs to refine attributes of the function being defined:
%
%                     'delayed'   : whether the function has delayed-evaluation semantics (default: 1)
%
%                     'hold'      : []/'all'/'first'/'rest'; whether to leave all/the first/the remaining arguments or no arguments of the function unevaluated 
%                                   before executing the function body (useful for higher-order functions), (default: [])
%
%                     'symbolic'  : whether the function being defined should not return unevaluated when receiving unevaluated expressions as arguments
%                                   (ignoring those arguments that fall under hold), for symbol-manipulating functions (0/1, default: 0)
%
%                     'add_impure': integrate the original expression into the function value (when evaluated), creating an impure expression, 0/1
%                                   if 0, the output expression may still be impure, e.g., when the function body introduces that field into the result
%
%                     'add_online': integrate an online expression into the function value (when evaluated); relevant only when the settings 
%                                   attribute is 'filter' or 'editing', may also be specified in exp_enddef
%                                   * 'passthrough': the function is skipped in the online processing
%                                   * 'endofbody':   the function is executed in the online processing
%                                     (using the function arguments as they appear at the end of the function's body, i.e. when exp_enddef is invoked), 
%                                   * otherwise: the supplied argument is directly used as the online expression
%
% Out:
%   Immediate-Eval     : whether the body of the function being defined shall be executed or skipped
%
% Advanced functionality:
%   The function's behavior can be controlled via dynamically scoped variables, as set by exp_block, exp_set and exp_setdelayed, or correspondent attributes.
%     These variables can be assigned values, such as 0/1, but may be arbitrary expressions
%      * fingerprint_check: toggle per-argument fingerprint checking (default: true) - may use @expression to refer to the argument expression
%      * fingerprint_create: toggle fingerprint creation (default: true) - may use @expression to refer to the unevaluated complete expression
%      * memoize: selectively enable memoization; this is a cell array of the form {location,expression,location,expression, ...}.
%                 location can be either 'disk' or 'memory', expression should evaluate to true/false, may use @expression to refer to the unevaluated 
%                 complete expression.
%
% See also:
%   exp_enddef(), exp_eval()
%
% Examples:
%     function result = myfunc(myarg,myotherarg,yetanotherarg)
%     if ~exp_begindef('somesetting'), return; end
%
%     do_any_computation;
%
%     exp_enddef;
%
%                                        Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                        2010-04-14

% read the current execution context
try
    throw; %#ok<LTARG> fastest way to get an exception
catch
    ctx = lasterror; %#ok<LERR>
    
    % check if the expression system shall be disabled (only applicable in filter/editing expressions)
    if any(strcmp(setting,{'filter','editing'})) && isequal(hlp_lookup_symbol(@disable_expressions,ctx),{1})
        % and return immediately, if so
        assignin('caller','exp_internal_context',struct('enabled',{0}, 'exec_ctx',{ctx}));
        immediate = 1;
        return;
    end
    
    % identify the caller
    callname = ctx.stack(2).name;
    callfile = ctx.stack(2).file;
    callfunc = str2func(callname);
    [codehash,inargs,outargs] = utl_fileinfo(callfile,callname);
    
    % whether we should evaluate immediately or not depends on whether we are being called by exp_eval()
    immediate = any(strcmp('exp_eval',{ctx.stack(3:min(4,length(ctx.stack))).name}));
end

%read the caller's function arguments, unfolding varargs along the way
try
    parts = {};
    for i=1:length(inargs)
        val = evalin('caller',inargs{i});
        if ~strcmp(inargs{i},'varargin')
            val = {val}; end
        parts = [parts val]; %#ok<AGROW>
    end   
catch,end

% parse the options
if isempty(varargin)
    % no custom attributes, fast path
    switch setting
        case 'symbolic'
            % standard symbolic expressions...
            if ~immediate       
                % compose & return a new expression from transformed arguments
                assignin('caller',outargs{1},struct('head',{str2func(callname)},'parts',{parts}));
            else
                % opt out of exp_enddef() processing
                assignin('caller','exp_internal_context',struct('enabled',{0}, 'exec_ctx',{ctx}));
            end
            return;
        case {'filter','editing'}
            % standard filter & editing expressions...
            opts = struct('symbolic',0, 'hold',{[]}, 'delayed',1,'add_impure',1,'argsteps',{{@check_dataset}}, ...
                'presteps',{[]}, 'poststeps',{{@add_online_exp}}, 'fingerprint_check', {[]}, 'fingerprint_create',{[]}, 'memoize',{[]});
            if strcmp(setting,'filter')
                opts.add_online = 'endofbody';
            else
                opts.add_online = 'passthrough';
            end            
    end
else
    % baseline attributes
    attribs = {'hold',[],'symbolic',0,'add_online',[],'argsteps',[],'presteps',[],'poststeps',[],'fingerprint_check',[],'fingerprint_create',[],'memoize',[]};
    % setting-dependent attributes
    switch setting        
        case 'filter'
            defaults = {'delayed',1,'add_impure',1,'add_online','endofbody','argsteps',{@check_dataset},'poststeps',{@add_online_exp}};
        case 'editing'
            defaults = {'delayed',1,'add_impure',1,'add_online','passthrough','argsteps',{@check_dataset},'poststeps',{@add_online_exp}};
        case 'symbolic'
            defaults = {'delayed',1,'add_impure',0,'symbolic',1,'hold','all','fingerprint_check',0,'fingerprint_create','passthrough','memoize',0};
        case 'default'
            defaults = {'delayed',1,'add_impure',0,'argsteps',{@check_fingerprint}};
        case 'annotate'
            defaults = {'delayed',0,'add_impure',1};
    end
    % slow path for everything else
    opts = hlp_varargin2struct(varargin, attribs{:}, defaults{:});
end

% the delayed option can override immediacy...
immediate = ~opts.delayed || immediate;

% compose a new expression from transformed arguments
expression = struct('head',{callfunc},'parts',{parts},'codehash',{codehash});

if immediate
    % purify the expression
    expression = utl_purify_expression(expression);
    % quick check if memoization is disabled
    if ~isequal(opts.memoize,0)
        % do a cache lookup
        [action,result] = utl_memoize_lookup(expression,opts.memoize);
        if strcmp(action,'return')
            % return the retrieved value
            assignin('caller',outargs{1},result);
            immediate = 0;
            return;
        elseif strcmp(action,'memoize')
            % remember to memoize the expression later, under the given id(s)
            memoize_id = result;
        else
            % do not memoize the expression
            memoize_id = [];
        end
    else
        memoize_id = [];
    end
    
    % begin measuring the evaluation time of the expression (used to determine the best cache location)
    eval_time = tic;
    
    % apply per-argument transformation steps
%     for step = opts.argsteps
%         parts = cellfun(@(arg)step{1}(arg,opts),parts,'UniformOutput',false); end
    
    % evaluate the function's arguments (recursively), if they are not held
    if ~isempty(parts) && ~any(strcmp(opts.hold,{'all','first'}));
        parts{1} = exp_eval(parts{1});
        if ~opts.symbolic && ~is_evaluated(parts{1})
            % cannot be evaluated: return the expression
            assignin('caller',outargs{1},expression);            
            immediate = false;
            return;
        end
    end
    if ~any(strcmp(opts.hold,{'all','rest'}))
        for i=2:length(parts)
            parts{i} = exp_eval(parts{i}); %#ok<AGROW>
            if ~opts.symbolic && ~is_evaluated(parts{i})
                % cannot be evaluated: return the expression
                assignin('caller',outargs{1},expression);
                immediate = false;
                return;
            end
        end
    end
            
    % create a context struct to keep track of the inputs, outputs, and attributes of the function throughout its evaluation
    context = struct('inargs',{inargs}, 'outargs',{outargs}, 'ws_input_pre',{struct()}, 'ws_output_pre',{struct()}, 'opts',{opts}, 'eval_time', {eval_time}, ...
        'expression_preeval',{expression}, 'expression_posteval', struct('head',{callfunc},'parts',{parts}), 'memoize_id',{memoize_id}, 'exec_ctx',{ctx}, 'enabled',{1});
    
    % re-derive the function's input & output workspace from the now evaluated parts, folding varargs
    for i=1:length(parts)
        if strcmp(inargs{i},'varargin')
            context.ws_input_pre.(inargs{i}) = parts(i:end);
            break;
        else
            context.ws_input_pre.(inargs{i}) = parts{i};
        end
    end
    
    % execute the prestep(s) on the workspaces
    for step = opts.presteps'
        context = step{1}(context); end
    
    % assign all input & output arguments for the function body's execution
    for fn=fieldnames(context.ws_input_pre)'
        assignin('caller',fn{1},context.ws_input_pre.(fn{1})); end
    for fn=fieldnames(context.ws_output_pre)'
        assignin('caller',fn{1},context.ws_output_pre.(fn{1})); end
    
    % store the context object...
    assignin('caller','exp_internal_context',context);
else
    % delayed evaluation: return the expression
    assignin('caller',outargs{1},expression);    
end


% argstep to check whether the value of impure expressions is consistent with the expressions themselves
function x = check_fingerprint(x,opts)
% get the fingerprint checking expression (either from the options or from the dynamic context) 
if isempty(opts.fingerprint_check)
    opts.fingerprint_check = exp_eval(@fingerprint_check); end
% this check can be selectively overridden with a block-scoped symbol @fingerprint_check, which may be an expression that involves @expression
%(the expression for which we are about to execute the check)
if is_impure_expression(x) && ~isequal(0,exp_eval(exp_releasehold(exp_replacerepeated(opts.fingerprint_check,{exp_rule(@expression,x)}))))
    if isfield(sig.bcilab,'fingerprint') && ~isequal(hlp_fingerprint(rmfield(x,'bcilab')),x.bcilab.fingerprint)
        error('expression is inconsistent with the attached value'); end
end



% argstep for 'filter'/'editing' functions, to check, fix up and warn about the value of inconsistent impure expressions
% (expressions referring to signals, in particular) and dataset values (without any notion of expressions)
function sig = check_dataset(sig,opts)
% get the fingerprint checking expression (either from the options or from the dynamic context) 
if isempty(opts.fingerprint_check)
    sym = hlp_lookup_symbol(@fingerprint_check);
    if ~isempty(sym)
        opts.fingerprint_check = sym{1}; end
end
% we only operate on signals
% this check can be selectively overridden with a block-scoped symbol @fingerprint_check, which may be an expression that involves @expression
% (the expression for which we are about to execute the check)
if all(isfield(sig,{'data','chanlocs'})) && ~isequal(0,exp_eval(utl_releasehold(utl_replacerepeated(opts.fingerprint_check,{exp_rule(@expression,sig)}))))
    
    if (~isfield(sig,'bcilab') || ~(~isfield(sig.bcilab,'fingerprint') || isequal(hlp_fingerprint(rmfield(sig,'bcilab')),sig.bcilab.fingerprint)))

        % the dataset has been imported from EEGLAB or has been edited with it: do an implict re-import using set_new()
        if isfield(sig,'bcilab') && isfield(sig.bcilab,'expression')
            sig = rmfield(sig.bcilab,'expression'); end
        sig = set_new(sig);
        % check for problems by scanning the history
        found_problem = 0;
        if isfield(sig,'history')
            % catch a few known to be {catastrophic,dangerous,subtle} operations in the imported dataset
            operations = { ...
                {{'pop_eegfilt','pop_runica'}, ...
                'The operation ''%s'' will neither allow to generate reliable predictions nor usable online models from the imported dataset.'}, ...
                {{'pop_averef','pop_interp','pop_reref','pop_resample','pop_rmbase','pop_subcomp'}, ...
                'The operation ''%s'' will not allow to generate online models from the imported dataset.'}, ...
                {{'pop_chansel','pop_rejchan','pop_rejchanspec','pop_select'}, ...
                ['The operation ''%s'' requires that the channels supplied during online processing (and their order) match those' ...
                ' in the present dataset.']}, ...
                };
            for l=1:length(operations)
                for op=operations{l}{1}
                    if strfind(sig.history,[op{1} '('])
                        warning('BCILAB:exp_begindef:dataset_messed_up',operations{l}{2},op{1});
                        found_problem = 1;
                    end
                end
            end
            if found_problem
                warning('BCILAB:exp_begindef:dataset_mess_recomm',['\nIt is recommended that these operations be executed from BCILAB''s ' ...
                    'processing palette,\nsince the operations implemented there are online-capable.\n\n']);
            end
        end
        if ~found_problem
            warning('BCILAB:exp_begindef:dataset_dirty',['The dataset has been edited or imported from EEGLAB; to guarantee ' ...
                'correctness of BCILAB''s\n performance estimates, ensure that the imported data had not previously been filtered\n' ...
                '(spatially, spectrally, or temporally).\n\n']);
        end
        warning('BCILAB:exp_begindef:memory_waste',['\nA copy of the data will be kept in memory for all derived datasets.\n' ...
            'To reclaim this memory, instead save the original edited data to disk using EEGLAB and load it via BCILAB''s io_loadset()\n\n']);
    end
end



% poststep to insert a .bcilab.online_expression field into the dataset;
% by default, use the arguments as they appear after execution of the function body (excluding inputs which have
% themselves an online_expression - these are substituted by their online expression)
function context = add_online_exp(context)
if strcmp(context.opts.add_online,'passthrough')
    % the function shall be skipped online
    online_expressions = {};
    % we need to the find the part among the original input which contains the online_expression
    for k=1:length(context.expression_posteval.parts)
        exp = context.expression_posteval.parts{k};
        if is_impure_expression(exp) && isfield(exp.bcilab,'online_expression')
            online_expressions{end+1} = exp.bcilab.online_expression; end %#ok<AGROW>
    end
    if isempty(online_expressions)
        error('this stage cannot be skipped, since it does have not any input signal');
    elseif length(online_expressions) > 1
        % make sure that they are all the same...
        if ~all(cellfun(@(e) exp_eval(exp_same(e,online_expressions{1})),online_expressions(2:end)))
            error('this stage cannot be skipped, since it has multiple mutually different input signals'); end
    end
    % and assign that to the first output
    final_expression = online_expressions{1};
elseif strcmp(context.opts.add_online,'endofbody')
    % the function will participate in online processing; we parse its input argument values and compose a new expression from these
    
    % first check if an intermediate argument has been deleted (which would be a case of serious misconduct, if the modified inputs 
    % shall be the inputs for the online expression)
    arg_deleted = 0;
    for k=1:length(context.inargs)
        argname = context.inargs{k};
        if ~isfield(context.ws_input_post,argname)
            arg_deleted = 1;
        elseif arg_deleted
            error('an intermediate argument was cleared from the input arguments; in this case, the online expression is undefined.');
        end        
    end
    
    % construct expression parts from the current input arguments, expanding varargins
    online_parts = {};
    for k=1:length(context.inargs)
        argname = context.inargs{k};        
        try            
            tmp = context.ws_input_post.(argname);
            if strcmp(context.inargs{k},'varargin')
                online_parts = [online_parts tmp]; %#ok<AGROW>
            else
                online_parts{end+1} = tmp; %#ok<AGROW>
            end
        catch, end
    end
    
    % go through the parts of the original expression and copy the original (but evaluated) online expressions from there, to be used as 
    % sub-expressions in place of the impure or changed values after running the function body
    for k=1:length(context.expression_posteval.parts)
        exp = context.expression_posteval.parts{k};
        if is_impure_expression(exp) && isfield(exp.bcilab,'online_expression')
            online_parts{k} = exp.bcilab.online_expression; end         %#ok<AGROW>
    end    
    
    % construct the final expression
    final_expression = struct('head',{context.expression_posteval.head},'parts',{online_parts});
    if length(context.outargs) > 1 && isfield(context.ws_output_post,context.outargs{2})
        % if a second output is present & assigned, we treat it as a state output, and include it in the final expression
        final_expression.state = context.ws_output_post.(context.outargs{2}); end
else
    % an online expression has been specified as an attribute; use that
    final_expression = context.opts.add_online;
end

% assign the expression to the .bcilab.online_expression field of the first output (-to-be)
if isfield(context.ws_output_post.(context.outargs{1}),'bcilab')
    context.ws_output_post.(context.outargs{1}).bcilab.online_expression = final_expression; end
