function exp_enddef(varargin)
% exp_enddef(Arguments...)
% End the definition of a function with delayed-evaluation semantics, see exp_begindef. Invokes plugin functions and handles memoization.
%
% In:
%   Arguments... : optional name-value pairs; second place to override some fields & behaviors, aside from exp_begindef
%                   'add_impure': see exp_begindef
%                   'add_online': see exp_begindef
%
% Examples:
%   see exp_begindef(), exp_eval()
%
% Notes:
%   The exp_enddef must be called in every code path that exits the function being defined (except for error conditions), i.e.
%   a premature return must be preceded by and exp_enddef;
%
%                                        Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                        2010-04-15

% retrieve & update the context structure
context = evalin('caller','exp_internal_context');
if ~context.enabled
    return; end

for i=1:length(context.inargs)
    try context.ws_input_post.(context.inargs{i}) = evalin('caller',context.inargs{i}); catch, end
end
for i=1:length(context.outargs)
    try context.ws_output_post.(context.outargs{i}) = evalin('caller',context.outargs{i}); catch, end
end
if ~isfield(context,'ws_input_post')
    context.ws_input_post = struct(); end
if ~isfield(context,'ws_output_post')
    context.ws_output_post = struct(); end

% read additional inputs
opts = hlp_varargin2struct(varargin);
for fn=fieldnames(opts)'
    context.opts.(fn{1}) = opts.(fn{1}); end

if context.opts.add_impure
    % if impure, incorporate the pure expression into the function result (if it's a struct)
    if isfield(context.ws_output_post,context.outargs{1}) && isstruct(context.ws_output_post.(context.outargs{1}))
        context.ws_output_post.(context.outargs{1}).bcilab.expression = context.expression_preeval; end
end

% run the poststep(s)
for s = 1:length(context.opts.poststeps)
    context = context.opts.poststeps{s}(context); end

% handle fingerprinting
if isempty(context.opts.fingerprint_create)
    sym = hlp_lookup_symbol(@fingerprint_create);
    if ~isempty(sym)        
        context.opts.fingerprint_create = sym{1}; end
end
if strcmp(context.opts.fingerprint_create,'passthrough')
    % no changes are being made to the fingerprint of the output
elseif isfield(context.ws_output_post,context.outargs{1}) && is_impure_expression(context.ws_output_post.(context.outargs{1}))
    if ~isequal(0,context.opts.fingerprint_create) && ~isequal(0,exp_eval(utl_replacerepeated(context.opts.fingerprint_create,{exp_rule(@expression,context.expression_preeval)}))) ...
        % fingerprint creation turned on: fingerprint the data, so we can later check the consistency of impure expression
        context.ws_output_post.(context.outargs{1}).bcilab.fingerprint = hlp_fingerprint(rmfield(context.ws_output_post.(context.outargs{1}),'bcilab'));
    else
        % fingerprint creation turned off: make sure that the fingerprint gets removed...
        if isfield(context.ws_output_post.(context.outargs{1}).bcilab,'fingerprint')
            context.ws_output_post.(context.outargs{1}).bcilab = rmfield(context.ws_output_post.(context.outargs{1}).bcilab,'fingerprint'); end
    end
end

% remember the computation time of the object
if isfield(context.ws_output_post,context.outargs{1}) && is_impure_expression(context.ws_output_post.(context.outargs{1}))
    context.ws_output_post.(context.outargs{1}).bcilab.computation_time = toc(context.eval_time); end

% assign the changed values in the workspace
for fn=fieldnames(context.ws_input_post)'
    assignin('caller',fn{1},context.ws_input_post.(fn{1})); end
for fn=fieldnames(context.ws_output_post)'
    assignin('caller',fn{1},context.ws_output_post.(fn{1})); end


% memoize the results (for the original expression)
if ~isempty(context.memoize_id) && isfield(context.ws_output_post,context.outargs{1})
    obj = context.ws_output_post.(context.outargs{1});
    % for each location... (currently: memory or disk)
    for c=1:length(context.memoize_id)
        id = context.memoize_id{c};
        utl_memoize_commit(obj,id);
    end
end
