function eeg = flt_clean_channels(eeg,min_corr,ignored_quantile,window_len,ignored_time,rereferenced,removed_channels)
% Out-Set= flt_clean_channels(In-Set,Min-Correlation,Window-Length,Max-Ignoredtime)
% remove channels with abnormal (uncorrelated) data from a (sufficiently high-passed) continuous dataset. Non-causal.
%
% This is an automated artifact rejection function which ensures that the dataset contains no channels that record complete trash.
% If channels with control signals are contained in the data, these are usually also removed.
%
% In:
%   In-Set           : continuous dataset, assumed to be appropriately high-passed (e.g. >0.5Hz or with a 0.5Hz - 2.0Hz transition band)
%
%   Min-Correlation  : minimum correlation between a channel and any other channel (in a window) below which the window is considered abnormal
%                      (default: 0.6)
%
%   Ignored-Quantile : quantile of the (sorted) correlation values that is ignored, including self-correlation (default: 0.1)
%
%   Window-Length    : length of the windows (in seconds) for which correlation is computed; ideally short enough to reasonably capture
%                      periods of global artifacts (which are ignored), but no shorter (for computational reasons) (default: 1)
% 
%   Max-Ignoredtime  : maximum time (in seconds) in the dataset, that is ignored (can contain arbitrary data without affecting the outcome)
%                      may also be a ratio of the total dataset length (default: 0.4)
%
%   Re-Referenced    : whether the operation should be run on re-referenced data (default: 0)
%
% Out:
%   Out-Set : EEG dataset with bad channels removed
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-07-06

if ~exp_begindef('filter') return; end;
if ~exist('min_corr','var') min_corr = 0.6; end
if ~exist('ignored_quantile','var') ignored_quantile = 0.1; end
if ~exist('window_len','var') window_len = 1; end
if ~exist('ignored_time','var') ignored_time = 0.4; end
if ~exist('rereferenced','var') rereferenced = 0; end

if ~exist('removed_channels','var')    
    if ignored_time > 0 && ignored_time < 1
        ignored_time = size(eeg.data,2)*ignored_time;
    else
        ignored_time = eeg.srate*ignored_time;
    end
    
    [C,S] = size(eeg.data);
    window_len = window_len*eeg.srate;
    wnd = 0:window_len-1;
    offsets = 1:window_len:S-window_len;
    W = length(offsets);    
    retained = 1:(C-ceil(C*ignored_quantile));

    % optionally subtract common reference from data
    if rereferenced
        X = eeg.data - repmat(mean(eeg.data),C,1);
    else
        X = eeg.data;
    end
    
    flagged = zeros(C,W);
    % for each window, flag channels with too low correlation to any other channel (outside the ignored quantile)
    for o=1:W
        sortcc = sort(abs(corrcoef(X(:,offsets(o)+wnd)')));
        flagged(:,o) = all(sortcc(retained,:) < min_corr);
    end
    % mark all channels for removal which have more flagged samples than the maximum number of ignored samples
    removed_channels = find(sum(flagged,2)*window_len > ignored_time);
end

% execute
eeg = pop_select(eeg,'nochannel',removed_channels);

exp_enddef;
