function [eeg,newstate] = flt_fir(eeg,oldstate,fspec,fmode,ftype,passripple,stopripple)
% [Result-Set,State] = flt_fir(Input-Set, State, Freqs, Mode, Type, Pass-Ripple,Stop-Ripple)
% Filter a continuous eeg set using a digital FIR lowpass/highpass/bandpass/bandstop/free-form filter [1].
%
% Digital FIR filters are computationally less efficient than IIR filters, but allow for somewhat more control. Specifically, FIR filters can not 
% "blow up" (diverge), even if extremely demanding frequency responses are implemented (e.g., drift removal). The computational cost of very low-frequency filters
% during online processing may be prohibitive, though. FIR filters can be designed with different phase (delay/distortion) behavior, depending on the desired
% application. Linear phase filters are the most commonly used ones, as they do not distort the data (which makes interpretation easier) but only delay it, and because they 
% are causal (i.e. can be used online). The delay can, however, easily be too large for certain time-sensitive online tasks (it is a function of the lower transition edge).
% Zero-phase filters are mostly interesting for visualization, as they neither delay nor distort the signal, but cannot be used for online applications, or within
% data analyses that estimate online application behavior. Minimum-phase filters can be used online, have very low latency, and can implement extreme frequency responses,
% but distort the signal. In that case, some assumptions about signal shape may turn invalid, and have to be revised given the filtered data.
% 
% In:
%   Input-Set    :   continuous eeg set to be filtered
%
%   State        :   previous filter state, as obtained by a previous execution of flt_fir on an immediately preceding dataset (default: [])
%
%   Freqs        :   frequency specification:
%                    * for a low/high-pass filter, this is: [transition-start, transition-end], in Hz
%                    * for a band-pass/stop filter, this is: [low-transition-start, low-transition-end, hi-transition-start, hi-transition-end], in Hz
%                    * for a free-form filter, this is a cell array of {[frequency, frequency, ...], [amplitude, amplitude, ...]}
%                      (frequencies in Hz, amplitudes usually between 0 and 1, not including frequency/amplitude pairs at 0 Hz and at the Nyquist frequency)
%                      Alternatively, it can also be a an array of {[frequency, frequency, ...], [amplitude, amplitude, ...], [ripple, ripple, ...]}
%                      where ripple is the amount of rippling, overriding the values specified as Pass-Ripple and Stop-Ripple.
%
%   Mode         :   filter mode, 'bp' for bandpass, 'hp' for highpass, 'lp' for lowpass, 'bs' for bandstop, 'ff' for free-form (default: 'bp')
%
%   Type         :   * 'minimum-phase' minimum-hase filter -- pro: introduces minimal signal delay; con: distorts the signal (default)
%                    * 'linear-phase' linear-phase filter -- pro: no signal distortion; con: delays the signal
%                    * 'zero-phase' zero-phase filter -- pro: no signal delay or distortion; con: can not be used for online purposes
%
%   Pass-Ripple  :   maximum allowed relative pass-band ripple; assumed to be in db if negative (default: -20)
%
%   Stop-Ripple  :   maximum allowed relative stop-band ripple; assumed to be in db if negative (default: -30)
%
% Out: 
%   Result-Set   :  filtered, continuous eeg set
%   State        :  state of the filter, after it got applied
%
% Example:
%   filtered = flt_fir(raw,[],[5 10 26 37]); % Apply a digital FIR band-pass filter to the raw data set.
%
% Notes:
%   The designed filter is optimal in terms of maximum absolute error (as opposed to mean square error). The Signal Processing Toolbox is required.
%
% References:
%   [1] A.V. Oppenheim and R.W. Schafer, "Digital Signal Processing",
%       Prentice-Hall, 1975.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-17  


if ~exp_begindef('filter') return; end
if ~exist('fmode','var') || isempty(fmode) fmode = 'bp'; end;
if ~exist('ftype','var') || isempty(ftype) ftype = 'minimum-phase'; end;
if ~exist('passripple','var') || isempty(passripple) passripple = -20; end;
if ~exist('stopripple','var') || isempty(stopripple) stopripple = -30; end;

if passripple < 0
    passripple = 10^(passripple/10); end
if stopripple < 0
    stopripple = 10^(stopripple/10); end

% phase 1: design the filter if necessary
if iscell(fspec) && length(fspec)==1
    % called with precomputed filter coefficients
    b = fspec{1};
else
    % create filter specification
    amp_table = struct('bp',[0 1 1 0],'bs',[1 0 0 1],'lp',[1 0],'hp',[0 1]);
    if ~strcmp(fmode,'ff')
        % look up the amplitude from the table
        fspec = {fspec,amp_table.(fmode)}; end
    if length(fspec) < 3
        % derive the rippling specification from the amplitudes and passripple/stopripple
        fspec{3} = stopripple + fspec{2}*(passripple-stopripple); end    
    if strcmp(ftype,'zero-phase')
        % filter is being applied twice: correct for that
        fspec{2} = sqrt(fspec{2}); end
    % design the filter as minimax-optimal
    c = firpmord(fspec{:}, eeg.srate, 'cell');
    b = hlp_microcache('fdesign',@firpm,max(3,c{1}),c{2:end});
    if strcmp(ftype,'minimum-phase')
        % minimze the phase
       [dummy,b] = rceps(b); end %#ok<ASGLU>
end

% phase 2: filter the data
sig = double(eeg.data)';
if isempty(oldstate) 
    % offline case; we need to prepend the signal with a mirror section of itself, to minimize start-up transients
    sig = [repmat(2*sig(1,:),length(b),1) - sig((length(b)+1):-1:2,:); sig]; 
    if strcmp(ftype,'zero-phase')
        % to get a zero-phase filter, we run the filter backwards first
        % reverse the signal and prepend it with a mirror section (to minimize start-up transients)
        sig = sig(end:-1:1,:); sig = [repmat(2*sig(1,:),length(b),1) - sig((length(b)+1):-1:2,:); sig];
        % run the filter
        sig = filter_fast(b,1,sig);
        % reverse and cut startup segment again
        sig = sig(end:-1:(length(b)+1),:);
    end
else
    % online case: check for misuses
    if strcmp(ftype,'zero-phase')
        error('zero-phase filters are non-causal and cannot be run online (or on continued data); use linear-phase or minimum-phase filters, or flt_iir.'); end
end

% apply the filter
[sig,newstate] = filter_fast(b,1,sig,oldstate);

% check if we need to cut off a data segment that we previously prepended
if isempty(oldstate)    
    sig(1:length(b),:) = []; end

% write the data back
eeg.data = sig';

% remember the precomputed filter coeffs
fspec = {b}; %#ok<NASGU>
exp_enddef;
