function eeg = flt_reconstruct(eeg, dict, varargin)
% Out-Set = flt_reconstruct(In-Set, Dictionary, Arguments...)
% Reconstruct the given data in a new (possibly overcomplete) basis.
%
% Sparse Reconstruction is an approach to re-representing a signal in a new basis, which may have (far)
% more basis vectors than the signal has channels; in these cases, a unique solution can only be found 
% if the additional assumption is incorporated that at any given time there is only a small (sparse) 
% set of basis vectors active. Possible use cases include re-representing EEG in an overcomplete ICA basis
% (i.e. more components than channels), or obtaining a sparse reconstruction of cortical surface area activation
% (in terms of a basis of surface patch activities).
%
% Note: currently too slow for EEG-sized signals; need a GPU solver for this.
%
% In:
%   In-Set       : epoched or continuous EEGLAB data set
%
%   Dictionary   : [#Channels x #BasisVectors] dictionary of basis vectors in terms of which
%                  the signal shall be reconstructed
%
%   Arguments... : optional name-value pairs, with possible names:
%                  'basisinfo': new channel locations, with one entry per basis vector
%                               can either be a cell array of channel names or a struct 
%                               with field 'labels' and possibly other fields (default: {'1','2','3',...})
%                  'lambda'   : regularization parameter; controls sparsity-accuracy tradeoff (default: 1)
%                  'variant'  : optimization method to be used, one of the following:
%                               'mackay': fast MacKay update rule for Sparse Bayesian learning
%                               'fastem': fast EM-base update rule for Sparse Bayesian learning (default)
%                               'slowem': slow ('traditional') EM-based update rule for Sparse Bayesian learning
%                               'focuss': FOCUSS algorithm for sparse recovery; uses the 'p'-valued quasi-norm L_p
%                               'l1': l1-norm sparse reconstruction (for additional options, see help NESTA_UP)
%                               'l2': minimum-energy reconstruction
%                  'transform': transform the data into thew new representation (default: 1)
%                               alternatively, a new field in the EEGLAB dataset would be created (yet to be agreed on)
%
%                  'p'        : p-valued quasi norm (for the FOCUSS mode) (default: 1)
%                  'maxiter'  : maximum number of iterations in the inner loop (default: 3000)
%                  'hyperprior': [BasisVectors x 1] vector of hyperprior values, gamma_0 in SBL (default: 0 = minimum-norm)
%                  'sigma'    : noise level estimate for the 'l1' variant (default: 0.1)
%                  'window_len': window length for the SBL variants, in seconds (default: 1)
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-03-28

if ~exp_begindef('filter') return; end;

% read parameters
if ~exist('dict','var') || isempty(dict) dict = eye(C); end
opts = hlp_varargin2struct(varargin, 'basisinfo',[], 'lambda',1, ...
    'variant','fastem', 'p',1, 'maxiter',3000, 'transform',1, 'hyperprior',[],'sigma',0.1,'Verbose',0,'window_len',1);

% reconstruct data
[C,S,T] = size(eeg.data);
eeg.data = reshape(eeg.data,C,[]);
switch opts.variant
    case 'l1'
        % use l1 reconstruction
        data = double(eeg.data);
        eeg.data = zeros(size(dict,2),S*T);
        try
            % use NESTA
            nesta_opts = rmfield(opts,setdiff(fieldnames(opts),{'xplug','U','normU','MaxIntIter','maxiter','TolVar','stopTest','TypeMin','Verbose','fid','errFcn','outFcn'}));
            dictnorm = norm(dict*dict');
            lambda = opts.sigma * sqrt(2*log(C));
            mufinal = 0.1*opts.sigma/dictnorm;
            for t=1:S*T
                eeg.data(:,t) = NESTA_UP(dict,[],data(:,t),lambda,dictnorm,mufinal,nesta_opts); end
        catch
            % use the CVX fallback
            N = size(dict,2);
            for t=1:S*T
                cvx_begin
                    variables X(N)
                    minimize(norm(X,1))
                    subject to
                        norm(dict*X - data(:,t)) <= opts.sigma;
                cvx_end
                eeg.data(:,t) = X;
            end
        end
    case {'mackay','fastem','slowem','focuss'}        
        % use sparse Bayesian learning
        mode = struct('mackay',0,'fastem',1,'slowem',2,'focuss',[3 opts.p]);
        if isempty(opts.hyperprior)
            opts.hyperprior = 0; end        
        % process the signal in overlapped windows
        window_len = opts.window_len*eeg.srate;
        wnd = 0:window_len-1;
        wnd_weight = repmat(hann(length(wnd))',C,1);
        offsets = 1 + floor(0:window_len/2:(S*T)-window_len);
        W = length(offsets);
        X = eeg.data;
        eeg.data = zeros(size(dict,2),S*T);
        for o=1:W
            S = X(:,offsets(o) + wnd).*wnd_weight;
            eeg.data(:,offsets(o)+wnd) = eeg.data(:,offsets(o)+wnd) + sparse_learning_fast(dict,S,opts.lambda,opts.maxiter,mode.(opts.variant),opts.hyperprior,0);
        end
    case 'l2'
        % use the pseudoinverse
        eeg.data = dict\eeg.data;
    otherwise
        error(['unsupported variant selected: ' opts.variant]);
end

if opts.transform
    % override data
    eeg.data = reshape(eeg.data,[],S,T);
    eeg.nbchan = size(eeg.data,1);
    % rewrite chanlocs
    if isempty(opts.basisinfo)
        eeg.chanlocs = struct('labels',cellfun(@num2str,num2cell(1:eeg.nbchan,1),'UniformOutput',false));
    elseif length(opts.basisinfo) == eeg.nbchan
        if isfield(opts.basisinfo,'labels')
            eeg.chanlocs = opts.basisinfo;
        elseif iscellstr(opts.basisinfo)
            eeg.chanlocs = pop_chanedit(struct('labels',opts.basisinfo),'lookup',[]);
        else
            error('unsupported format for the ''basisinfo'' parameter');
        end        
    else
        error('length of ''basisinfo'' parameter does not match number of basis vectors');
    end
else
    error('transform is currently required to be 1');
end

exp_enddef;
