function [eeg,newstate] = flt_resample(eeg, oldstate, srate)
% [Result-Set,State] = flt_resample(Input-Set, State, Sample-Rate)
% Changes the sampling rate of a given dataset.
%
% Resampling [1] is usually applied to reduce the computational cost of certain methods, by running them on a signal sampled at a lower rate. Other reasons to resample
% data are to save memory, to improve a method's robustness (by reducing its number of parameters) or to align and/or unify signals originally sampled at different rates. 
% Resampling has two drawbacks that require consideration: first, it may under some conditions create frequency artifacts (called aliasing), which can be amplified 
% by the adaptive information-extracting nature of later stages in most BCI paradigms; in the worst case, this can lead to scientifically invalid conclusions. Second,
% it introduces a delay into the signal (usually on the order of 10-100ms), which may be prohibitive in very time-critical online scenarios. If absolutely necessary,
% this delay can be mitigated by the use of some specialized (signal-distorting and/or more aliasing-prone) resamplers.
%
% In:
%   Input-Set   :   continuous dataset to be resampled
%
%   State       :   state of a previous filter invocation
%
%   Sample-Rate :   the new sample rate
%
% Out:
%   Result-Set  :   a resampled dataset
%
%   State       :   state after application of the filter
%
% Examples:
%   eeg = flt_resample(eeg, 100); -- resample to 100 Hz
%
% References:
%  [1] Proakis, J., and Manolakis, D. "Digital Signal Processing: Principles, Algorithms and Applications."
%      Macmillan Publishing Company, 1992.
%
%                   Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                   2010-03-28

if ~exp_begindef('filter') return; end
if ~exist('oldstate','var') oldstate = []; end  
if ~exist('srate','var') srate = eeg.srate; end

if size(eeg.data,3) > 1
    error('flt_resample can not be applied to epoched data.'); end

if srate ~= eeg.srate
    % we need to resample
    
    % compute a tight upper bound on the number of samples needed from the past & future of the base stream (= our delay & overlap); 
    % this is imposed by resample(); the 10 is the default value being used
    [p,q] = rat(srate/eeg.srate, 0.0001);
    N = ceil(10*max(1,q/p));
    
    % we keep track of a state struct, which is formatted as:
    % .buf: dataset(past N samples of the original dataset ("base stream")),
    % .bufofs: 0-based base stream sample number of the first sample in the state dataset,
    % .delivered: 0-based base stream sample number up to which we have already delivered
    
    if isempty(oldstate)
        % no previous state, so this is the offline case (in the online case we have by design always valid previous state)
        if size(eeg.data,2) < N
            error(['the dataset must have at least ' num2str(N) ' samples to allow for resampling.']); end        
        % make up a sufficient amount of back-extrapolated state to minimize startup transients, using a reflection method
        % its sample range (0-based, into to the base stream, is -N:-1), so the bufofs is -N
        % we consider this portion of the data already delivered (so it won't be emitted subsequently), so delivered is -1
        oldstate = struct('buf',exp_eval(set_new(eeg, 'data',repmat(2*eeg.data(:,1),1,N) - eeg.data(:,(N+1):-1:2), ...
            'xmin',0,'icaact',[],'event',eeg.event([]),'urevent',eeg.urevent([]),'epoch',[],'reject',[],'specdata',[],'specicaact',[])), 'bufofs',-N, 'delivered',-1);
    end
    
    % compute what parts of the output data we are going to retain
    %   0-based index range into the base data set (in eeg)
    base_range = oldstate.bufofs + (0:(size(eeg.data,2)+size(oldstate.buf.data,2)-1));
    %   sampling positions into the base data set that we hit after we resample
    out_range = base_range(1) + (0:ceil(length(base_range)*p/q)-1) * q/p;
    %   those indices in the out_range that are not already delivered and valid
    retain = find(out_range > oldstate.delivered & out_range < base_range(end)-N);    
    
    % prepend past state (now we know the base stream sample range of eeg.data)
    eeg = pop_mergeset(oldstate.buf,eeg);
    
    % update the state (the last N samples of the data, for later continuation, and the updated base stream sample number)
    newstate = struct('buf',pop_select(eeg,'point',size(eeg.data,2)-N+1:size(eeg.data,2)),'bufofs',oldstate.bufofs+size(eeg.data,2),'delivered',out_range(retain(end)));    
    
    % resample the data, with boundary functionality deactivated
    bounds = find(strcmp({eeg.event.type},'boundary')); 
    subs = repmat({'__disabled_'},length(bounds),1); [eeg.event(bounds).type] = subs{:};    
    eeg = pop_resample(eeg, srate);
    subs = repmat({'boundary'},length(bounds),1); [eeg.event(bounds).type] = subs{:};    

    % retain only those portions of the output data that are valid (i.e. have not been delivered already and are not incomplete)
    eeg = pop_select(eeg,'point',[retain(1) retain(end)]);
end

oldstate = [];
exp_enddef;
