function [eeg,state] = flt_standardize(eeg,state,window_len)
% [Result-Set,State] = flt_standardize(Input-Set, State, Window-Length)
% Standardize a continuous EEG set causally to have a per-channel standard deviation of 1.
%
% Standardization ensures that per-channel data that normally can have any variance (e.g. due to varying conductivity, 
% amplifier settings, etc.) is approximately normally distributed over datasets and time periods. This is usually necessary when 
% learning and running models across sessions and subjects. It should not be applied before other artifact-rejection steps, as 
% these steps usually take relative signal power into account. It is important to make the standardization window long enough that 
% it does not factor out changes in signal power that one is interested in.
%
% In:
%   Input-Set    :   continuous eeg set to be filtered
%
%   State        :   previous filter state, as obtained by a previous execution of flt_iir on an immediately preceding dataset (default: [])
%
%   Window-Len   :   length of the window based on which normalization shall be performed, in seconds (default: 30)
%
% Out: 
%   Result-Set   :  standardized, continuous eeg set
%
%   State        :  state of the filter, after it got applied
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-17

if ~exp_begindef('filter') return; end
if ~exist('window_len','var') || isempty(window_len) window_len = 30; end;

% number of data points for our normalization window
N = window_len * eeg.srate;
B = ones(1,N)/N;

% make up prior state if necessary
if isempty(state)
    if size(eeg.data,2) < N+1
        error(['For standardization, the dataset needs to be longer than the set data window length (currently ' num2str(window_len) 'seconds).']); end
    % filter conditions & constant overall data offset (for better numerical accuracy; this is unrelated to the running mean)
    state = struct('ord1',[],'ord2',[],'offset',sum(eeg.data,2)/size(eeg.data,2));
    % prepend a made up data sequence
    eeg.data = [repmat(2*eeg.data(:,1),1,length(B)) - eeg.data(:,(length(B)+1):-1:2) eeg.data];
    prepended = true;
else
    prepended = false;
end

% get raw data X and running mean E[X]
X = bsxfun(@minus,double(eeg.data),state.offset);
[X_mean,state.ord1] = filter_fast(B,1,X,state.ord1,2);
% get squared data X^2 and running squared mean E[X^2]
X2 = X.^2;
[X2_mean,state.ord2] = filter_fast(B,1,X2,state.ord2,2);
% compute running std deviation sqrt(E[X^2] - E[X]^2)
X_std = sqrt(X2_mean - X_mean.^2);

% compute standardized data
eeg.data = (X - X_mean) ./ X_std;

% trim the prepended part if there was one
if prepended
    eeg.data(:,1:length(B)) = []; end

exp_enddef;
