function res = io_loadset(filename, channels, srate)
% Set = io_loadset(Filename (,Channels,Sampling-rate))
% Loads/imports a dataset from disk, across all formats supported by EEGLAB
% optionally, data can be loaded in reduced form to help conserve memory where applicable.
%
% In:
%   Filename      :   name of the file; platform-independent path preferred.
%   Channels      :   optional: names of channels to be loaded
%   Sampling-Rate :   optional: rate at which the data should be sampled
%
% Out:
%   Set :   dataset in EEGLAB format
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-03-28

if ~exp_begindef('filter') return; end
if ~exist('channels','var') 
    channels = []; end
if ~exist('srate','var') 
    srate = []; end

% otherwise construct it
filename = env_translatepath(filename);
disp(['loading ' filename '...']);

% dispatch to the appropriate loader / importer
[base,name,ext] = fileparts(filename);
switch lower(ext)
    case '.set'
        % EEGLAB dataset
        res = pop_loadset('filepath',[base filesep], 'filename', [name ext]);
        if ~isfield(res,'bcilab') || ~isfield(res.bcilab,'online_expression')
            disp('The loaded EEGLAB set is lacking an online expression; assuming it contains unfiltered data.')
            disp('If it contains filtered data, however, BCI models derived from it will likely not be online-capable.');
        end
    case '.vhdr'
        % BrainProducts Vision Recorder file
        if ~isempty(channels) && isnumeric(channels)
            res = pop_loadbv([base filesep],[name ext], [], channels);
        else
            res = pop_loadbv([base filesep],[name ext]);
        end
        % remove non-standard event fields
        res.event = rmfield(res.event,intersect({'code','bvtime','channel'},fieldnames(res.event)));
    case '.raw'
        try
            % EGI continuous raw data        
            res = pop_readegi(filename);
        catch
            % ... or ERPSS raw
            res = pop_read_erpss(filename,srate);
        end
    case '.sna'
        % Snapmaster SNA
        res = pop_snapread(filename);
    case '.cnt'
        % Neuroscan CNT
        try
            res = pop_loadcnt(filename);
        catch
            % ... or ANT EEProbe CNT
            res = pop_loadeep(filename,'triggerfile','on');
        end
    case '.avr'
        % ANT EEProbe average file
        res = pop_loadeep_avg(filename);
    case '.eeg'
        % Neuroscan EEG
        if exist('channels','var') && ~isempty(channels) && isnumeric(channels)
            res = pop_loadeeg([name ext],[base filesep],channels);
        else
            res = pop_loadeeg([name ext],[base filesep]);
        end
    case {'.bdf','.edf'}
        % BioSEMI BDF/EDF
        try
            if exist('channels','var') && ~isempty(channels) && isnumeric(channels)
                res = pop_biosig(filename,channels);
            else
                res = pop_biosig(filename,[]);
            end
        catch
            % backup variant
            res = pop_readbdf(filename);
        end
    case '.mat'
        % BrainVision Analyzer Matlab file
        res = pop_loadbva(filename);
    case '.ds'
        % CTF folder
        res = pop_ctf_read(filename);
    case '.rdf'
        % ERPSS data
        res = pop_read_erpss(filename,srate);
    case '.asc'
        % INStep ASC
        res = pop_loadascinstep(filename);
    case '.m4d'
        % 4D pdf file
        res = pop_read4d(filename);
    otherwise
        % Miscellaneous formats
        try
            % try File-IO
            if ~isempty(channels) && isnumeric(channels)
                res = pop_fileio(filename,'channels',channels);
            else
                res = pop_fileio(filename);
            end
        catch
            fileio_error = lasterror; %#ok<LERR>
            try
                % try BioSig
                if ~isempty(channels) && isnumeric(channels)
                    res = pop_biosig(filename,channels);
                else
                    res = pop_biosig(filename,[]);
                end
            catch
                biosig_error = lasterror; %#ok<LERR>
                try
                    % try to load it as a BCI2000 ASCII file
                    res = pop_loadbci(filename,srate);
                catch
                    env_handleerror(fileio_error);
                    env_handleerror(biosig_error);
                    error('Unknown data format; please check for plugins at www.sccn.ucsd.edu/eeglab/plugins/');                    
                end
            end
        end
end

% if no channel locations present, try to look them up
if ~all(isfield(res.chanlocs,{'X','Y','Z'})) || all(cellfun('isempty',{res.chanlocs.X}))
    % ... using the 10-20 system
    try
        res = pop_chanedit(res,'lookup','standard-10-5-cap385.elp');
        fraction_found = 1-mean(cellfun('isempty',{res.chanlocs.X}));
        % if only a small fraction of channel names were found
        if fraction_found < 0.25
            % check if it's the SCCN's custom cap
            locdb = readlocs(env_translatepath('resources:/sccn-cap-128.elp'));
            [found,idx_in_locdb,idx_in_res] = intersect({locdb.labels},{res.chanlocs.labels});
            if length(found)/length(res.chanlocs) > fraction_found
                % we found more channel names in that cap: copy the fields from there
                for fname = fieldnames(locdb)'
                    [res.chanlocs(idx_in_res).(fname{1})] = locdb(idx_in_locdb).(fname{1}); end
            end
        end
        % optimize the head center
        res = pop_chanedit(res,'eval','chans = pop_chancenter( chans, [],[]);');
    catch
    end
end

% describe it (note: the @rawdata expression indicates that this stage produces raw data...)
if isfield(res,'bcilab') && isfield(res.bcilab,'online_expression')
    exp_enddef('add_online',res.bcilab.online_expression)
else
    exp_enddef('add_online',@rawdata);
end

% and reduce the data if not already done so in the loader
if exist('channels','var') && ~isempty(channels)
    res = set_selchans(res,channels); end

if exist('srate','var') && ~isempty(srate)
    res = flt_resample(res,srate); end
