function model = ml_traingmm(trials, targets, clusters, varargin)
% Model = ml_traingmm(Trials, Targets, Clusters, Options...)
% Learn a probabilistic non-linear predictive model using Gaussian Mixture Models.
% 
% Gaussian Mixture Models [1] are well-known generative statistical models that can be used for clustering, classification, and density estimation, whereas
% the provided implementation is primarily targeted at classification. The basics assumption of Gaussian mixture models is that the data (in each class)
% is a mixture of (relatively few) approximately gaussian-distributed distinct clusters. The optimization methods used for finding these clusters are usually
% very fragile (compared to, e.g., SVMs, where this problem is bypassed) and give reasonable results when the data is in fact a collection of distinct clusters, 
% each of which deviates not too much from being gaussian, and when there are sufficient samples per cluster, w.r.t. to the total dimensionality of the data.
% These assumptions are usually only fulfilled when the data of one (or more) class(es) was derived from multiple distinct situations (e.g. walking, driving,
% waiting) and when either a large number of trials is available, or the dimensionality of the feature space is relatively low, i.e. the features come from
% a an effective adaptive feature-extraction stage.
%
% Several variants are supplied, among others "conventional" approaches using, e.g., expectation maximization, which require that the number of clusters is
% approximately known in advance. These are implemented via the GMMBAYES library [2]. Further, there are "advanced" implementations, where the number
% of clusters is not needed in advance, but learned from the data, usually using a variant of the Dirichlet prior, which are implemented using the VDPGM [3] 
% library.
%
% A benefit of GMMs is that the produced probability are of the highest quality among all methods implemented by the toolbox. For example, data points which 
% are not in any of the previously learned classes typically receive equal probabilities under all classes (which cannot be expected from logistic regression
% or relevance vector machines). However, due to their restrictions, gaussian mixture models should be viewed as rather specialized classifiers, which should
% only be relied on when the assumptions are fulfilled.
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%
%   Clusters     : expected/maximum number of clusters (required for some variants)
%                  default: 3
%
%   Options  : optional name-value parameters to control the training details:
%              'variant': one of several variants, including
%                        "nonparametric" methods:
%                        'avdp': accelerated variational Dirichlet process, ignores Clusters
%                        'vdp' : variational Dirichlet process iterative, using a per-weight prior, 
%                                   more or less uninformative, but numerically more robust than the 'vb' variant
%                        'bj'  : Blei & Jordan
%                        'cdp' : collapsed Dirichlet prior
%                        'csp' : collapsed stick-breaking
%                        conventional parametric methods:
%                        'vb'  : variational Bayes (exact # clusters)
%                        'em'  : expectation-maximization (exact # cluster)
%                        'fj'  : Figueiredo-Jain (max. # clusters)
%                        'gem' : greedy expectation-maximization (max. # clusters)
%              
%              'scaling': pre-scaling of the data (see hlp_findscaling for options) (default: 'std') 
%
% Out:
%   Model   : a multi-class gaussian mixture model; 
%             classes indicates the class labels which the model predicts
%
% Examples:
%   model = ml_traingmm(trials,labels): learn a gaussian mixture classifier
%   
% See also:
%   ml_predictgmm()
%
% References:
%   [1] Trevor Hastie and Robert Tibshirani, "Discriminant Analysis by Gaussian Mixtures"
%       Journal of the Royal Statistical Society. Series B (Methodological), Vol. 58, No. 1 (1996), pp. 155..176
%   [2] Nico Vlassis and A. Likas, "A greedy EM algorithm for gaussian mixture learning."
%       Neural Process. Lett. 15, 1 (2002), 77?87.
%   [3] Kenichi Kurihara, Max Welling and Nikos Vlassis "Accelerated Variational Dirichlet Mixture Models"
%       Advances in Neural Information Processing Systems 19 (NIPS 2006).
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-05

% read parameters
if ~exist('clusters','var') || isempty(clusters) clusters = 3; end
opts = hlp_varargin2struct(varargin, 'variant','avdp', 'scaling','std');

% scale the data
sc_info = hlp_findscaling(trials,opts.scaling);
trials = hlp_applyscaling(trials,sc_info);

% identify and remap the classes
classes = unique(targets);
% remap target labels to 1..k (bias is added later)
targ = targets;
for c=1:length(classes)
    targets(targ==classes(c)) = c; end

switch opts.variant
    case {'avdp','vdp'}
        for c=1:length(classes)
            model.data{c} = trials(targets==c,:);
            model.class{c} = vdpgm(model.data{c}',feval(['mkopts_' opts.variant]));
        end
    case {'bj','bjrnd','cdp','csb','vb'}
        for c=1:length(classes)
            model.data{c} = trials(targets==c,:);
            model.class{c} = vdpgm(model.data{c}',feval(['mkopts_' opts.variant], clusters));
        end
    case 'em'
        model.class = gmmb_create(trials, targets, 'EM', 'components', clusters);
    case 'fj'
        model.class = gmmb_create(trials, targets, 'FJ', 'Cmax', clusters);
    case 'gem'
        model.class = gmmb_create(trials, targets, 'GEM', 'Cmax', clusters);
    otherwise
        error('unknown variant specified');
end

model.classes = classes;
model.sc_info = sc_info;
model.variant = opts.variant;