function model = ml_trainhkl(trials, targets, lambdas, varargin)
% Model = ml_trainhkl(Trials, Targets, Lambdas, Options...)
% Learn a sparse non-linear predictive model using Hierarchical Kernel Learning.
%
% Hiararchical Kernel Learning [1,2] is a non-linear, kernel-based (as in Support Vector Machines) learning algorithm which can be used both for regression
% and classification. It can give probabilistic outputs (similar to logistic regression). The defining property of HKL is that it can select relevant 
% features out of a large number of potentially many irrelevant ones (similar to LASSO), whereas these features are selected according to non-linear criteria,
% which makes HKL uniquely powerful in selecting relevant portions of the data. 
%
% The method is relatively slow due to the need for regularization, but on the other hand has the highest chance of learning relevant structure in
% very large numbers of dimensions (among the provided machine learning methods). For these reasons, it is not the best method to start with,
% as others allow to evaluate data and to iterate through parameters at a much faster rate. It is, however, one of the best methods to try in order to
% get the best possible results, and also to check whether data on which other methods have failed contains any information of interest at all.
% 
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%
%   Lambdas      : vector of regularization parameters to optimize for
%                  default: 10.^[1:-.5:-8]; note: if this is specified as a vector instead of a search() expression,
%                  ml_trainhkl will internally optimize the parameter using a more efficient procedure; search() can nevertheless
%                  be used instead, to get finer control over cross-validation (e.g. safety margins, etc.)
%
%   Options  : optional name-value pairs to control the training details:
%              'loss'  : either 'logistic' for classification or 'squared' for regression (default: 'logistic')
%              'kernel': type of kernel to use; can be one of the following;
%                        'hermite': hermite polynomial kernel (default)
%                        'spline' spline kernel 
%                        'gauss-hermite-full':  hermite expansion of hte Gaussian kernel (finite order)
%                        'anova': ANOVA kernel
%                        'polynomial': polynomial kernel (not recommended)
%              'params': vector of kernel parameters; if empty, chosen according to the author's recommendations (default: [])
%                        see hkl.m for a detailed explanation of this parameter
%              'mingap': result precision; expressed as duality gap (default: 1e-3)
%              'gapeta': smoothing of the reduced problem (default: 1e-3)
%              'maxactive': maximum number of selected kernel (default: 100)
%              'display': display diagnostic outputs (default: 0)
%              'data_normalization': either 'center' (for centering) or 'scale' (for centering and standardization) (default: 'scale')
%              'kernel_normalization': 'center' (for centering) 'scale' (for centering and standardization), or 'scale-root' (default: 'scale-root')
%              'alpha': dual parameter, see [2]
%              'eta': kernel weights, see [2]
%              'b': constant term, see [2]
%              'memory_cache': size of the memory cache; should be large enough to fill most of the free RAM (default: 1e9)
%              'k': foldness of an internal cross-validation, to compute the optimal regularization parameter (default: 5)
%
% Out:
%   Model   : a HKL model...
%
% See also:
%   ml_predicthkl()
%
% References:
%   [1] F. Bach. "Exploring Large Feature Spaces with Hierarchical Multiple Kernel Learning." 
%       Advances in Neural Information Processing Systems (NIPS), 2008.
%   [2] F. Bach. "High-Dimensional Non-Linear Variable Selection through Hierarchical Kernel Learning." 
%       Technical report, HAL 00413473, 2009.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-04

% read parameters
if ~exist('lambdas','var') || isempty(lambdas) lambdas = 10.^[1:-.5:-8]; end
opts = hlp_varargin2struct(varargin, 'loss','logistic', 'kernel','hermite', 'params',[], 'mingap',1e-3, 'gapeta',1e-3, ...
    'maxactive',400, 'display',0, 'data_normalization','scale', 'kernel_normalization','scale-root', 'memory_cache',1e9, 'k',5);

if iscell(targets)
    targets = targets{1};
    disp('note: hkl does not support weighted learning.');
end


% identify and remap the classes, if necessary
classes = unique(targets);
if strcmp(opts.loss,'logistic')
    if length(classes) > 2
        % in this case we use the voter...
        model = ml_trainvote(trials,targets,@ml_trainhkl,@ml_predicthkl,lambdas,varargin{:});
        return;
    else
        % remap targets
        targets(targets == classes(1)) = 0;
        targets(targets == classes(2)) = 1;
    end
end

% select default kernel parameter, if necessary
if isempty(opts.params)
    switch opts.kernel
        case 'polynomial', opts.params = [3 .1 4];
        case 'hermite', opts.params = [.5 3 .1 4];
        case 'gauss-hermite', opts.params = [1 .05 3 .1 .5];
        case 'gauss-hermite-full', opts.params = [1 .05 3 .1 .5 30];
        case 'anova', opts.params = [.05 .1 4 30];
        case 'spline', opts.params = [.1 4 40];
        otherwise
            error('unknown kernel type specified');
    end
end

% learn
warning off MATLAB:illConditionedMatrix
args = hlp_struct2varargin(opts,'restrict', {'mingap','gapeta','maxactive','display','data_normalization','kernel_normalization','alpha','eta','b','memory_cache','conjgrad'});
if length(lambdas)>1
    % find the best lambda by randomized CV
    if ~opts.display
        evalc('[model.outputs,model.model,dum1,dum2,model.bestlambda] = hkl_kfold(opts.k,''same'',trials,targets,lambdas,opts.loss,opts.kernel,opts.params,args{:});');
    else
        [model.outputs,model.model,dum1,dum2,model.bestlambda] = hkl_kfold(opts.k,'same',trials,targets,lambdas,opts.loss,opts.kernel,opts.params,args{:}); %#ok<ASGLU>
    end
else
    % compute the model directly for the given lambda
    [model.outputs,model.model] = hkl(trials,targets,lambdas,opts.loss,opts.kernel,opts.params,args{:});
    model.bestlambda = 1;
end

model.classes = classes;
model.opts = opts;