function model = ml_trainlda(trials, targets, lambda, varargin)
% Model = ml_trainlda(Trials, Targets, Lambda, Options...)
% Learn a linear predictive model by (regularized) Linear Discriminant Analysis.
%
% LDA is one of the simplest and oldest learning algorithms, originally introduced by Fisher in [1]. Its basic assumption is that the 
% data originates from two classes (extended to more via 1-vs-1 voting), where the data in each class is distributed in the feature space 
% according to a normal distribution. The shape of the distribution is assumed to be identical for both classes, and the relative weight/prior 
% probability of each class is (by default) assumed to be identical, as well. An example would be data with features generated as a weighted
% sum of some latent variables which take on different values between conditions (but identical values within each condition), and which are superimposed
% with gaussian noise (which is identically distributed across conditions, e.g. from a large sum of independent random variables).
%
% Despite its simplicity, LDA assumes more structure in the data than is usually necessary (namely a certain normal distribution per class), and sometimes,
% more than what can be satisfactorily learned from the data, so that, even when the assumptions are fulfilled, the method is not guaranteed to give the
% optimal result. The estimation of the per-class covariance matrix is a notoriously data-hungry (and especially outlier-sensitive) step, and the main
% weakness of standard LDA. 
%
% ml_trainlda offers three advanced covariance estimators, with different trade-offs to mitigate these problems. The variants 'shrinkage' and 'independence' 
% each introduce a regularization parameter [2] which controls the complexity of the estimated matrix, and which need to be selected in a parameter search
% (which is orders of magnitude more time-consuming). The 'auto' variant computes the degree of regularization analyically in closed form, and is therefore
% both fast and (in some sense) optimal (but more restricted than 'independence') [3]. When enough trials are available, full covariance matrices are 
% learned, but the less trials are given, the more the covariance estimates degrade to spherical (though well-formed) ones. Auto is the default for lda.
%
% While regularization can automatically control the complexity of a classifier, it is not a panacea which allows to add arbitrarily many features,
% since with each additional feature, the amount of structure (here:correlation) that can be captured in the remaining ones gets reduced.
%
% Within the toolbox, LDA is one of the bread-and-butter classifiers, and is worth trying in every reasonably simple classification task, especially
% for its speed. The major problem with LDA compared to other available classifiers is that it easier to break it with outliers than, for example,
% support vector machines or logistic regression. Another weakness is that the outputs are relatively primitive probability estimates, in contrast
% to, for example, logistic regression (somewhat better) or relevance vector machines (clearly better).
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%
%   Lambda       : optional regularization parameter, reasonable range: 0:0.1:1, greater is stronger
%                  requires that the regularization mode is set to either 'shrinkage' or 'independence' (default: [])
%           
%   Options  : optional name-value parameters to control the training details:
%              'regularization' -> 'shrinkage': covariance shrinkage, depends on lambda 
%                                  'independence': feature independence, depends on lambda
%                                  'auto': analytical covariance shrinkage, lambda is ignored (default)
%              'weight_bias' -> 0/1, take unequal class priors into account for bias calculation
%                               default: 0
%              'weight_cov' -> 0/1, take unequal class priors into account for covariance calculation
%                              default: 0
% Out:
%   Model   : a linear model; w is the linear weights, b is the bias; classes indicates the class labels which the model predicts
%
% Examples:
%   model = ml_trainlda(trials,labels): learn a simple LDA model
%   
% See also:
%   ml_predictlda()
%
% References:
%   [1] Fisher, R. "The use of multiple measurements in taxonomic problems."
%       Annals Eugen. 7 (1936), 188, 179.
%   [2] Friedman, J. "Regularized discriminant analysis." 
%       Journal of the American Statistical Association 84, 405 (1989), 175, 165.
%   [3] O. Ledoit and M. Wolf, "A well-conditioned estimator for large-dimensional covariance matrices"
%       J Multivar Anal, 88(2): 365-411, 2004.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-03

opts = hlp_varargin2struct(varargin, 'weight_bias',0, 'weight_cov',0, 'regularization','auto');
if ~exist('lambda','var') 
    lambda = []; end

% find the class labels
classes = unique(targets);
if length(classes) > 2
    % learn a voting arrangement of models...
    model = ml_trainvote(trials, targets, @ml_trainlda, @ml_predictlda, lambda, varargin{:});   
else    
    % estimate distribution of each class...
    for c = 1:length(classes)
        % get mean and covariance
        X = trials(targets==classes(c),:);
        n{c} = size(X,1);
        mu{c} = mean(X);
        if strcmp(opts.regularization,'auto')
            sig{c} = cov_shrink(X);
        else
            sig{c} = cov(X);
            if ~isempty(lambda)
                % lambda-dependent regularization
                if strcmp(opts.regularization,'independence')
                    sig{c} = (1-lambda)*sig{c} + lambda * diag(diag(sig{c}));
                elseif strcmp(opts.regularization,'shrinkage')
                    sig{c} = (1-lambda)*sig{c} + lambda*eye(length(sig{c}))*abs(mean(diag(sig{c})));
                else
                    error('unknown regularization mode');
                end
            end
        end
    end
    
    ns = fastif(opts.weight_cov,n,{1 1});
    nb = fastif(opts.weight_bias,n,{1 1});
    % do the math
    mu_both = (mu{1}*nb{2} + mu{2}*nb{1}) / (nb{1}+nb{2});    
    sig_both = (sig{1}*ns{1} + sig{2}*ns{2}) / (ns{1}+ns{2});
    w = (mu{2} - mu{1}) / sig_both;
    w = w / (mu{2}*w' - mu_both*w');
    model = struct('w',{w}, 'b',{mu_both*w'}, 'classes',{classes});
end
