function model = ml_trainsvmlinear(trials, targets, cost, varargin)
% Model = ml_trainsvmlinear(Trials, Targets, Cost, Options...)
% Learn a linear predictive model by Support Vector Machines.
%
% Linear support vector machines are nowadays some of the most frequently used and versatile linear classifiers [1]. They can handle a large number of
% feature efficiently and are usually extremely robust (if regularized well) or fast (when not regularized). The main differences between the provided
% variants are computational, and allow to select the most efficient model for the data (in terms of #trials/#features). All variants except for 'matlab'
% are implemented using the LIBLINEAR package [2], the 'matlab' variant servers as fallback (in case LIBLINEAR is not available for some platform). 
% Support vector machines do not naturally produce good probability estimates (although some approach exists, like in the LDA case).
%
% in typical uses of the toolbox, linear SVMs are not as frequently used as logreg or lda, primarily due to the need for proper regularization, 
% which is often prohibitively costly. For quick tests, the method can be used without regularization, however. When extremely many feature or
% trials are used, linear SVMs are likely more useful.
%
% In:
%   Trials       : training data, as in ml_train
%
%   Targets      : target variable, as in ml_train
%
%   Cost         : regularization parameter, reasonable range: 2.^(-5:2:15), greater is stronger
%                  default: 1
%
%   Options  : optional name-value pairs to control the training details:
%              'variant': one of several linear SVM variants
%                       'dual'    : L2-regularized L2-loss support vector classification (dual), usually preferred (default)
%                       'primal'  : L2-regularized L2-loss support vector classification (primal), can be faster than dual
%                       'crammer' : Crammer and Singer approach to multi-class support vector classification,
%                                   as alternative to the voted classification employed in the other variants
%                       'sparse'  : L1-regularized L2-loss support vector classification, gives sparse results but
%                                   will likely not be better unless the solution is truly sparse.
%                       'l1loss'  : L2-regularized L1-loss support vector classification (dual), rarely needed
%                       'matlab'  : MATLAB implementation (using CVX); equivalent to 'dual', but slower
%
%              'eps'    : desired accuracy (default: [], meaning LIBLINEAR's default)
%
%              'bias'   : include a bias variable (default: 1)
%
%              'scaling': pre-scaling of the data (see hlp_findscaling for options) (default: 'std') 
%
%              'quiet'  : quiet mode - no outputs (default: 1)
%
% Out:
%   Model   : a linear model; 
%             classes indicates the class labels which the model predicts
%             sc_info is the scaling info
%
% Examples:
%   model = ml_trainsvmlinear(trials,labels,1): learn a linear SVM model
%   
% See also:
%   ml_predictsvmlinear()
%
% References:
%   [1] P. S. Bradley and O. L. Mangasarian. "Massive data discrimination via linear support vector machines."
%       Optimization Methods and Software, 13:1-10, 2000. 
%   [2] R.-E. Fan, K.-W. Chang, C.-J. Hsieh, X.-R. Wang, and C.-J. Lin. "LIBLINEAR: A library for large linear classification" 
%       Journal of Machine Learning Research 9(2008), 1871-1874.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-04

% read parameters
if ~exist('cost','var') || isempty(cost) cost = 1; end
opts = hlp_varargin2struct(varargin, 'variant','dual', 'eps',[], 'bias',1, 'scaling','std', 'quiet',1);

% identify and remap the classes (to 0..k)
classes = unique(targets);
if strcmp(opts.variant,'matlab') 
    % MATLAB version
    if length(classes) > 2
        % in this case we use the voter...
        model = ml_trainvote(trials,targets,@ml_trainsvmlinear,@ml_predictsvmlinear,cost,varargin{:});
        return;
    end
    [n,f] = size(trials);
    % scale the data
    sc_info = hlp_findscaling(trials,opts.scaling);
    trials = hlp_applyscaling(trials,sc_info);
    % remap targets
    targets(targets == classes(1)) = -1;
    targets(targets == classes(1)) = +1;    
    % solve
    cvx_begin
        variables w(f) b xi(n)
        minimize 1/2*sum(w.*w) + cost*sum(xi)
        subject to
            targets.*(trials*w + b) >= 1 - xi;
            xi >= 0;
    cvx_end
    model = struct('w',w,'b',b);
else
    % scale the data
    sc_info = hlp_findscaling(trials,opts.scaling);
    trials = hlp_applyscaling(trials,sc_info);        
    % remap targets
    targ = targets;
    for c=1:length(classes)
        targets(targ==classes(c)) = c-1; end
    
    % check whether we are using the fallback variant if LIBLINEAR
    fallback = ~isempty(strfind(which('train'),'pre7.3'));
    if fallback && strcmp(opts.variant,'sparse')
        error('Sparse classifiers are not supported in LIBLINEAR prior to MATLAB 7.3 (due to a breaking change in the sparse array MEX API).'); end
    
    % train a model
    variant2num = struct('dual','1','primal','2','crammer','4','l1loss','3','sparse','5');
    model = train(targets,sparse(double(trials)),[ ...
        '-s ' variant2num.(opts.variant) ...
        ' -c ' num2str(cost) ...
        fastif(~isempty(opts.eps),[' -e ' num2str(opts.eps)],'') ...
        fastif(opts.bias,[' -B ' num2str(opts.bias)],'') ...
        fastif(opts.quiet && ~fallback,' -q ','') ...
        ]);
end
% incorporate the scaling coefficients into the model
model.sc_info = sc_info;
model.classes = classes;
model.opts = opts;
