function model = ml_trainsvmperf(trials, targets, cost, varargin)
% Model = ml_trainsvmperf(Trials, Targets, Cost, Options...)
% learn a linear or non-linear predictive model by Support Vector Machines, as implemented in the SVMperf package.
%
% SVMperf [1] is a compehensive package providing linear and non-linear (kernel) Support Vector Machines [2] to do classification.
% Among the provided SVM implementations, it is the preferred method whenever more than just mis-classification rate is reqired as optimization
% criterion (more precisely: as loss function). 
%
% Support Vector Machines are available for various machine learning tasks (SVMperf supports classification, SVMlight additionally supports 
% regression and ranking), and are generally an excellent (fast and/or robust, depending on regularization choice) solution. Kernel SVMs are a special
% variant which allow for state-of-the-art non-linear learning (depending on the kernel type, the kernel parameter needs to be searched, 
% in addition to the cost parameter). The major drawback of support vector machines is that they do not produce useful probability outputs, which restricts
% their use in decision-making setups. The other drawback is the need for regularization, which can be complicated and time-consuming (e.g. incurring 
% nested cross-validation).
% 
% When the data is known to require a non-linear classifier (or regressor), Support Vector Machines, Relevance Vector Machines (ml_trainrvm) and 
% Hierarchical Kernel Learning (ml_trainhkl) are currently the only methods provided in the toolbox, and their major difference is whether 
% they give probabilistic outputs and how much regularization they require (and how well they are able to pick out relevant information in high-dimensional
% feature spaces).
%
% SVMperf supports a variety of kernels (rbf being the most useful one in most cases) a variety of advanced loss functions, and various computational 
% methods/variants that primarily differ in speed (depending on data size, number of features, etc.) [3,4,5,6]. SVMperf is not free for non-commercial uses.
%
% In:
%   Trials   : training data, as in ml_train
%
%   Targets  : target variable, as in ml_train
%
%   Cost     : regularization parameter, reasonable range: 2.^(-5:2:15), greater is stronger
%              default: 0.01
%
%   Options  : optional name-value parameters to control the training details
%               'loss_function': Loss function to use (0..5,10)
%                                * 0/'ZO'    Zero/one loss: 1 if vector of predictions contains error, 0 otherwise.
%                                * 1/'F1'    F1: 100 minus the F1-score in percent.
%                                * 2/'Err'   Errorrate: Percentage of errors in prediction vector.
%                                * 3/'PRBEP' Prec/Rec Breakeven: 100 minus PRBEP in percent.
%                                * 4/'Prec'  Prec@k: 100 minus precision at k in percent.
%                                * 5/'Rec'   Rec@k: 100 minus recall at k in percent.
%                                * 10/'ROC'  ROCArea: Percentage of swapped pos/neg pairs (i.e. 100 - ROCArea).
%               'method': choice of structural learning algorithm (0..9), see code below for details
%               'variant': CPSP variant for sparse kernel training (0,1,2,4, if method is 9), see code below for details
%               'slacknorm': L-norm to use for slack variables (1,2)
%               'loss_rescaling': Rescaling method to use for loss (1,2)
%               'sparse_basis': number of basis functions for sparse kernel approximation (e.g. 500)
%               'restarts': number of restarts during sparse kernel approximation (0+)
%              kernel parameters:
%               'kernel': type of kernel function (lin,poly,rbf,sig,user) (default: 'lin')
%               'd': parameter d in polynomial kernel (int)
%               'g': parameter gamma in rbf kernel (float); should be searched, e.g., using search(-15:2:6)
%               's': parameter s in sigmoid/poly kernel (float)
%               'r': parameter c in sigmoid/poly kernel (float)
%               'u': parameter of user-defined kernel (string)
%              misc options:
%               'eps': tolerance (e.g., 0.1)
%               'bias': bias present? (0,1, default:1)
%               'pr_k': k in Precision/Recall loss functions
%               'shrinking_heuristic': whether shrinking heuristic is used (0,1)
%               'quiet': quiet mode (0,1)
%               'scaling': pre-scaling, see hlp_findscaling (default: 'std')
%
% Out:
%   Model   : a linear model; 
%             classes indicates the class labels which the model predicts
%             sc_info is the scaling info
%
% Examples:
%   model = ml_trainsvmperf(trials,labels,1): learn a SVM model
%
% Notes:
%   uses the SVM-struct learning module: SVM-perf, V3.00, 15.07.2009
%     includes SVM-struct V3.10 for learning complex outputs, 14.08.08
%     includes SVM-light V6.20 quadratic optimizer, 14.08.08
%     SVM-perf, SVM-struct, SVM-light copyright (C) Thorsten Joachims, thorsten@joachims.org
%   
% See also:
%   ml_predictsvmperf()
%
% References:
%   [1] T. Joachims, "A Support Vector Method for Multivariate Performance Measures", 
%       Proceedings of the International Conference on Machine Learning (ICML), 2005.
%   [2] Schlkopf, B., and Smola, A. "Learning with Kernels: Support Vector Machines, Regularization, Optimization, and Beyond"
%       (Adaptive Computation and Machine Learning). The MIT Press, Dec. 2001.
%   [3] T. Joachims, "Training Linear SVMs in Linear Time", 
%       Proceedings of the ACM Conference on Knowledge Discovery and Data Mining (KDD), 2006.
%   [4] T. Joachims, Chun-Nam Yu, "Sparse Kernel SVMs via Cutting-Plane Training",
%       Proceedings of the European Conference on Machine Learning (ECML), 2009.
%   [5] I. Tsochantaridis, T. Joachims, T. Hofmann, and Y. Altun, "Large Margin Methods for Structured and Interdependent Output Variables", 
%       Journal of Machine Learning Research (JMLR), Vol. 6(Sep):1453-1484, 2005.
%   [6] T. Joachims, T. Finley, Chun-Nam Yu, "Cutting-Plane Training of Structural SVMs", 
%       Machine Learning Journal, to appear.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-04

% read parameters
opts = hlp_varargin2struct(varargin, ...
    ... % --- optimization options ---
    'loss_function', 2, ...      % Loss function to use. (default 2)
    ...                          %   * 0/'ZO'    Zero/one loss: 1 if vector of predictions contains error, 0 otherwise.
    ...                          %   * 1/'F1'    F1: 100 minus the F1-score in percent.
    ...                          %   * 2/'Err'   Errorrate: Percentage of errors in prediction vector.
    ...                          %   * 3/'PRBEP' Prec/Rec Breakeven: 100 minus PRBEP in percent.
    ...                          %   * 4/'Prec'  Prec@k: 100 minus precision at k in percent.
    ...                          %   * 5/'Rec'   Rec@k: 100 minus recall at k in percent.
    ...                          %   * 10/'ROC'  ROCArea: Percentage of swapped pos/neg pairs (i.e. 100 - ROCArea).
    'method', 9, ...             % choice of structural learning algorithm (default 9):
    ...                          %   * 0: n-slack algorithm described in [5]
    ...                          %   * 1: n-slack algorithm with shrinking heuristic
    ...                          %   * 2: 1-slack algorithm (primal) described in [6]
    ...                          %   * 3: 1-slack algorithm (dual) described in [6]
    ...                          %   * 4: 1-slack algorithm (dual) with constraint cache [6]
    ...                          %   * 9: custom algorithm in svm_struct_learn_custom.c
    'variant',2, ...             % Use CPSP algorithm for sparse kernel training (if method is 9)
    ...                          %   * 0: do not use CPSP algorithm from [4]
    ...                          %   * 1: CPSP using subset selection for preimages via 59/95 heuristic (see [4])
    ...                          %   * 2: CPSP using fixed point search (see [4], RBF kernel only) (default)
    ...                          %   * 4: CPSP using fixed point search with starting point via 59/95 heuristic (see [4], RBF kernel only)
    'slacknorm', 1, ...          % L-norm to use for slack variables. Use 1 for L1-norm, use 2 for squared slacks. (default 1)
    'loss_rescaling', 2, ...     % Rescaling method to use for loss. 1: slack rescaling, 2: margin rescaling (default 2)
    'sparse_basis',500,...       % Specifies the number of basis functions to use for sparse kernel approximation (default 500)
    'restarts',0, ...            % Number of times to recompute the sparse kernel approximation and restart the optimizer. (default 0)
    ... % --- kernel options ---
    'kernel',0, ...              % type of kernel function:
    ...                          %   * 0/'lin': linear (default)
    ...                          %   * 1/'poly': polynomial (s a*b+c)^d
    ...                          %   * 2/'rbf': radial basis function exp(-gamma ||a-b||^2)
    ...                          %   * 3/'sig': sigmoid tanh(s a*b + c)
    ...                          %   * 4/'user': user defined kernel from kernel.h        
    'd',[], ...                  % parameter d in polynomial kernel (int)
    'g',[], ...                  % parameter gamma in rbf kernel (float)
    's',[], ...                  % parameter s in sigmoid/poly kernel (float)
    'r',[], ...                  % parameter c in sigmoid/poly kernel (float)
    'u',[], ...                  % parameter of user-defined kernel (string)
    ... % --- misc options ---
    'quiet', 1, ...              % verbosity level
    'scaling', 'std', ...        % scaling mode (see hlp_findscaling)
    'eps',0.1, ...               % epsilon: allow that tolerance for termination criterion (default 0.1)
    'bias', [], ...              % bias enabled? (0/1, default:1, only for lin kernel)
    'pr_k',0, ...                % fraction of positive examples to use as value of k for
    ...                          %   Prec@k and Rec@k. 0 indicates to use (0.5 * #pos) for
    ...                          %   Prec@k and (2 * #pos) for Rec@k. #pos is the number of
    ...                          %   positive examples in the training set. (default 0)
    'shrinking_heuristic', 1 ... % Selects whether shrinking heuristic is used in the custom algorithms for linear kernel and errorrate loss. (default 1)
     );

if ~exist('cost','var') || isempty(cost) cost = 0.01; end
                            
% scale the data
sc_info = hlp_findscaling(trials,opts.scaling);
trials = hlp_applyscaling(trials,sc_info);

% find the class labels
classes = unique(targets);
if length(classes) > 2
    % in this case we use the voter
    model = ml_trainvote(trials,targets,@ml_trainsvmperf,@ml_predictsvmperf,cost,varargin{:});
else
    
    % remap target labels to -1,+1
    targets(targets==classes(1)) = -1;
    targets(targets==classes(2)) = +1;
    
    % build the command
    command = ['-c ' num2str(cost)];
    command = [command fastif(opts.quiet,' -v 0 -y 0',' -v 1 -y 1')];
    command = [command ' -p ' num2str(opts.slacknorm)];
    command = [command ' -o ' num2str(opts.loss_rescaling)];
    if ischar(opts.loss_function)
        mapping = struct('ZO',0,'F1',1,'Err',2,'PRBEP',3,'Prec',4,'Rec',5,'ROC',10);
        command = [command ' -l ' mapping.(opts.loss_function)];
    else
        command = [command ' -l ' num2str(opts.loss_function)];
    end
    command = [command ' -w ' num2str(opts.method)];
    command = [command ' -e ' num2str(opts.eps)];
    if ischar(opts.kernel)
        mapping = struct('lin',0,'poly',1,'rbf',2,'sig',3,'user',4);
        command = [command ' -t ' mapping.(opts.kernel)];
    else
        command = [command ' -t ' num2str(opts.kernel)];
    end
    if ~isempty(opts.d) command = [command ' -d ' num2str(opts.d)]; end
    if ~isempty(opts.g) command = [command ' -g ' num2str(opts.g)]; end
    if ~isempty(opts.s) command = [command ' -s ' num2str(opts.s)]; end
    if ~isempty(opts.r) command = [command ' -r ' num2str(opts.r)]; end
    if ~isempty(opts.u) command = [command ' -u ' opts.u]; end
    if ~isempty(opts.bias) command = [command ' --b ' num2str(opts.bias)]; end
    command = [command ' --p ' num2str(opts.pr_k)];
    if opts.method == 9 command = [command ' --i ' num2str(opts.variant)]; end
    command = [command ' --k ' num2str(opts.sparse_basis)];
    command = [command ' --r ' num2str(opts.restarts)];
    command = [command ' --s ' num2str(opts.shrinking_heuristic)];
    
    % run the command
    model = svmperflearn(trials,targets,command);
    model.sc_info = sc_info;
    model.classes = classes;
end