function [prediction, measure, stats, target] = bci_predict(model, data, varargin)
% [Prediction, Loss, Statistics, Target] = bci_predict(Model, Dataset, Options...)
% Let a model make predictions of cognitive state on a dataset (usually with known target values), yielding one prediction per trial in the data, 
% as well as the average empirical loss, if the dataset has an attached target variable.
%
% bci_predict is used to obtain and/or evaluate the outputs of a predictive model on a recorded (usually raw) dataset. Predictions are made for every
% trial in the dataset, where trials are derived from the data in the same way as was used when the model was originally calibrated. For example, if the model
% was calibrated w.r.t. events of certain types in the calibration set (using the 'events' parameter of the paradigm), trials will be extracted for these 
% same events in the supplied dataset, as well. In the typical case where a target variable can be derived from the dataset in this way (e.g. via the 
% 'event' clause passed to bci_train - see bci_train), then the predictions of the model are compared with the defined target values, and an average loss
% is computed (and returned as the second output). This loss measure has the same meaning as in bci_train, and can be customized via the 'metric' parameter,
% either as one of the predefined losses (described in detail in machine_learning/ml_calcloss), or as a user-supplied custom function. By default, the most 
% appropriate loss is automatically selected depending on the types of target and prediction values (mis-classification rate for categorical target values, 
% mean-square error for continuous target values, etc.), so that user intervention is rarely needed.
%
% For advanced use, the actual predictions for every trial in the data are returned, as well as the defined target values in the dataset (if available).
% The format of the targets depends on the procedure by which they were assigned to the data (e.g., when the 'events' clause was used to assign target values
% to certain events, targets is a numerical vector of class numbers, e.g. [1 2 2 1 4 1 3 2 3 2 1] for some hypothetical 4-class dataset with 11 trials).
% The format of the predictions depends on the format of the targets (e.g., whether they were categorical values or continuous values, usually the former),
% and the capabilities of the 'learner' component of the paradigm. A detailed exposition of the possibilities is given in machine_learning/ml_predict. 
%
% Almost all learners produce discrete probability distributions for categorical targets, so this is the format of the predictions in 90% of the 
% situations. A sequence of discrete probability distributions (one per trial) as produced by bci_predict is formatted as a MATLAB cell array with the 
% three entries {'disc', Probabilities, Classes}. 'disc' is the format tag of the predictions and indicates discrete probabilities, Probabilities is an array of
% probabilities for each trial and class, with # of trials rows and # of classes columns, i.e. it has size [NxC], and Classes is a column vector which 
% specifies the desired target value for each of the C classes, in the same order as the rows in Probabilities. For this reason, the expected target value
% for any trial is pred{2}*pred{3}, and the most likely target value for any trial can be obtained as pred{3}(argmax(pred{2}')), if pred are the predictions
% return by bci_predict.
%
% In addition, further statistics are returned, which may contain additional values (e.g. false positive rate in the case of binary classification).
%
% Example:
%
%   % load a calibration dataset
%   calib = io_loadset('datasets/john/gestures.eeg')
%   % train a model (see bci_train for an explanation of this stage)...
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_csp}, 'events',{'left-imag','right-imag','rest'})
%   
%   % load a test dataset
%   testdata = io_loadset('datasets/john/gestures2.eeg')
%
%   % and apply the model on that dataset to get predictions and a loss estimate...
%   [predictions,loss2,stats2,targets] = bci_predict(model,testdata);
%
% The application of the model results in predictions for every of its trials, an empirical average loss measure, statistics, and the original target
% values for every trial in the test data. This is under the assumption that the file gestures2 contains markers 'left-imag', 'right-imag', 'rest', just 
% like the first dataset, so that a loss (deviation from desired outcomes) can be computed. Note that applying a model to its own calibration dataset gives
% overly optimistic results, and is usually scientifically invalid.
%
%
% A less common use case is to exchange the default preprocessing that is applied by a paradigm by some alternative preprocessing (for example,
% one which slighty deviates from the model's settings, or a manually implemented one, or the by the preprocessing of an entirely different model).
% To this end, not the raw data is passed to bci_predict, but instead data that was manually processed, for example using bci_preproc (see bci_preproc).
% In the follwing example, a test dataset is loaded (as in the previous sample), but it is manually preprocessed using the model's defined preprocessing,
% with re-customized options (here: 'events', to define how trials are to be extracted). We assume that the data contains events of the type 'user-action' whenever 
% the user performs a (not further known) action, such as one of the imagined hand gestures.
%  
%   testdata = io_loadset('datasets/john/unknown_gestures.eeg')
%   testdataproc = bci_preproc(testdata, model, 'events',{'user-action'})
%   predictions = bci_predict(model, testdataproc, 'process',0);
%
%   disp('the user performed the following actions: ' num2str(predictions{2}(argmax(predictions{3}')))]);
%
% In this case, we obtain predictions for every trial of the data (this time relative to events of the type 'user-action'), whereas the the predictions 
% will still be either 1,2, or 3, since the predictive part of the model has not been changed. Further, since the activity of the user was not known a priori, 
% a meaningful loss cannot be computed. Note that if the preprocessing of the model contains filters that are adaptively tuned to the data (such as ICA), this 
% strategy will give unexpected results, because bci_preproc re-runs the preprocessing (i.e. readapts the filters) for the test data, which gives intermediate values
% that are not expected by the rest of the model. 
%
% For these reasons, the recommended way to obtain predictions for a dataset at the times of arbitrary events, or in arbitrary intervals, is to use onl_stream, 
% which streams the given raw dataset into the model and obtains predictions at the desired time points. bci_predict should only be used to evaluate the performance
% of models on dataset where the desired outcomes are known and encoded in the same way as in the respective calibration dataset of the used model (i.e. for pure
% offline analyses). See onl_stream for further details.
%
%
% In:
%   Model      : a detector model, as produced by bci_train
%   Dataset    : a raw or fully preprocessed dataset from which to derive predictions
%   Options... : optional name-value pairs to control the function's behavior:
%                   'metric'  : evaluation metric to be employed, can be one of the followign: 
%                               * function handle: a custom, user-supplied loss function; receives target data in the first 
%                                 argument and prediction data in the second argument; each can be in any format that can be 
%                                 produced by ml_predict (but can be expected to be mutually consistent).
%                                 shall return a real number indicating the summary metric over all data, and optionally 
%                                 additional statistics in a struct
%                               * string: use ml_calcloss, with 'metric' determining the loss type
%                               * default/empty: use 'mcr','mse','nll','kld', depending on supplied target & prediction data formats
%
%                   'target'  : a function to derive the target variable from the data, for evaluation; the allowed format is anything 
%                               that may be output by ml_predict (default: @set_gettarget)
%                   'process' : whether to process & treat the dataset as raw data (0/1, default: 1)
%
% Out:
%   Prediction : the target variable as predicted by the model for every index in the dataset
%   Loss       : a measure of the average loss of the model, w.r.t. to the target variable (as derived by the target function), computed by the metric
%   Stats      : additional statistics, as produced by the metric
%   Target     : original target variable, as determined by the target function
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-05-24

% read arguments
opts = hlp_varargin2struct(varargin, 'target',@set_gettarget, 'metric',@(T,P) ml_calcloss([],T,P), 'process',1);
% string arguments are considered to be variants of the default metric
if isempty(opts.metric) || ischar(opts.metric) || (iscell(opts.metric) && all(cellfun(@ischar,opts.metric)))
    opts.metric = @(T,P)ml_calcloss(opts.metric,T,P); end
if ~has_stats(opts.metric)
    opts.metric = @(T,P)add_stats(opts.metric(T,P)); end

if opts.process
    % transform the dataset it using the model's online expression
    data = exp_eval(exp_replacerepeated(model.bcilab.online_expression,{exp_rule(@rawdata,data)}),1); end
% evaluate, using memoization and with fingerprinting turned off
data = utl_block({exp_rule(@fingerprint_check,0),exp_set(@fingerprint_create,0),exp_set(@memoize,{'disk',1})}, data);

% compute predictions
prediction = model.bcilab.prediction_function(data,'prediction','model',model);
try
    % try to derive the target variable
    target = opts.target(data);
    % and evaluate the performance, given the metric
    [measure,stats] = opts.metric(target,prediction);
catch
    % a target variable is not necessarily present
    target = [];
    measure = NaN;
    stats = struct();
end
stats.target = target;
stats.prediction = prediction;
stats.expression = struct('head',{@bci_predict}, 'parts',{{model,data.bcilab.expression,varargin{:}}});

% test whether the given metric supplies stats or not
function result = has_stats(metric)
try 
    [x,y] = metric([],[]);  %#ok<NASGU,ASGLU>
    result = true;
catch
    e = lasterror; %#ok<LERR>
    result = isempty(strmatch(e.identifier,{'MATLAB:TooManyOutputs','MATLAB:maxlhs','MATLAB:unassignedOutputs'},'exact'));
end

% add stats to the result of some metric
function [result,stats] = add_stats(result)
stats.value = result;
