function onl_feed(X)
% onl_feed(Data)
% Feed a block of raw data into the buffer.
%
% The function onl_feed is one of four functions in BCILAB's online processing Application Programming Interface (online API). The four functions are 
% onl_setup, onl_loaddetector, onl_feed, and onl_predict. For a complete example of using the online interface, see onl_setup. onl_feed is used to feed 
% a new block of data into an online stream (currently the global one). The data block must have the same number of channels as the the number of channels
% that were specified during the setup of the online stream, and the data must be sampled at the rate specified in the online stream. It is understood that
% no intermediate samples are missed when data is fed into the stream; If samples are in fact missed, zeros (or repeated values) should be fed to bridge the lost
% data portions. The feeding and processing of data in BCILAB is completely independent of real (wall-clock) time, and can therefore happen at any rate, 
% including in real time. Any output of the online system (obtained via onl_predict) refers to the most recent sample that was previously fed.
%
% In:
%   Data : [#Channels x #Samples] matrix of raw data
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-04-03

global bcilab;
if ~isfield(bcilab,'onl_buffer')
    error('onl_setup() must be called before onl_feed(), to set up the format of the data stream'); end

try
    if isempty(X)
        return; end
    
    % make sure the format of X is correct
    [C,S] = size(X);
    if ~isnumeric(X) || ndims(X) ~= 2
        error('onl_feed expects a [#Channels x #Samples] matrix'); end
    if ~isempty(bcilab.onl_buffer.data) && size(bcilab.onl_buffer.data,1) ~= C
        error('Number of channels in the supplied block does not match the number of channels in the buffer.'); end
    if length(bcilab.onl_buffer.chanlocs) ~= C
        error('Number of channels in the supplied block does not match the number of channel names in the buffer.'); end
    
    % append the block to the buffer
    bcilab.onl_buffer.data = [bcilab.onl_buffer.data X];
    [bcilab.onl_buffer.nbchan,bcilab.onl_buffer.pnts,bcilab.onl_buffer.trials] = size(bcilab.onl_buffer.data);
    bcilab.onl_buffer.xmax = bcilab.onl_buffer.xmax + S/bcilab.onl_buffer.srate;
    % smax is a field that is added (and fully maintained) by the online system; requires that filters do not resample the data.
    bcilab.onl_buffer.smax = bcilab.onl_buffer.smax + S; 
    if ~isfield(bcilab.onl_buffer,'bcilab') || ~isfield(bcilab.onl_buffer.bcilab,'online_expression')
        bcilab.onl_buffer.bcilab.online_expression = @dummy; end
catch
    disp('An exception occured inside onl_feed:');
    env_handleerror(lasterror); %#ok<LERR>
end