function result = para_dal_lf(data, mode, varargin)
% Result = para_dal_lf(Input-Data, Operation-Mode, Options...)
% Paradigm for slow cortical potentials, using the Dual-Augmented Lagrange method for trace-norm regularized logistic regression.
%
% The DAL-LF paradigm is, like para_windowmeans, a general method for operating on slow cortical potentials. It is a special case 
% of a more general framework described in [1] (using only its "first-order" detector); the general variant, which can in addition capture oscillatory
% processes, is further explained in para_dal. DAL is the name of the optimization method, and not an accepted or recognized name for BCI paradigms 
% using it (but is used here for the lack of a better name).
%
% The paradigm does not make a clear distinction between signal processing, feature extraction and machine learning, unlike most others, but instead 
% is a jointly optimized mapping from raw signal (epoch) to probabilistic prediction, using an efficient regularized optimization method (further 
% detailed in [2]). The method usually out-performs the windowed means paradigm, and in addition does not require any user parameters aside 
% from the epoch limits and lowpass filtering band, and is therefore one of the most useful BCI paradigms. The major drawback is the required computation
% time (and in some cases, the required memory -- which can be ameliorated by reducing the sampling rate of the data) due to the need for regularization.
% For this reason, it is a good strategy to first run the paradigm without regularization to get a ball-park estimate of the attainable accuracy, and only run
% the complete regularization when it makes sense.
% 
% Just like the windowed means paradigm, DAL-LF is applicable to a wide range of event-related and non-event-related scenarios, some of which are listed in 
% para_windowmeans.
%
% Example: Consider the goal of predicting whether a person perceives a fixated on-screen item as being unexpected (and/or erroneous, non-rewarding) 
% or not. A calibration dataset for this task could be annotated with an event for every gaze fixation made by the user (obtained from an eye tracker) 
% while reading short on-screen text fragments which are either semantically correct or incorrect. The two event types which identify the conditions sare 
% 'corr' and 'err'. From the literature [4,5], it can be assumed that these events should be accompanied by a characteristic slow cortical potential in the EEG,
% which allows to infer the condition. The 'learner' parameter will be specified as the default (relatively fine-grained) search over possible DAL regularization 
% parameter values.
%
%   calib = io_loadset('datasets/john/reading-errors.eeg')
%   [loss,model] = bci_train({'data',calib,'paradigm',@para_dal_lf}, 'epoch',[0 0.8], 'events',{'corr','err'}, 'spectrum',[0.1 15], ...
%       'learner',{'dal',search(2.^(-5:0.75:10))});
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'lowpass':  spectral filter (default: {[2 14]})
%                     * 'srate':   sampling rate, in Hz (default: 50)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [-1.5 1.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'normalizers': normalization parameters (default: [-0.25 -0.25])
%                     * 'learner': learning function to be used on the features; (default: 'dal')
%                                  note: use {'dal',search(2.^(-6:0.5:10)} here to run a parameter search
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References:
%  [1] Ryota Tomioka and Klaus-Robert Mueller, "A regularized discriminative framework for EEG analysis with application to brain-computer interface",
%      Neuroimage, 49 (1) pp. 415-432, 2010. 
%  [2] Ryota Tomioka & Masashi Sugiyama, "Dual Augmented Lagrangian Method for Efficient Sparse Reconstruction", 
%      IEEE Signal Proccesing Letters, 16 (12) pp. 1067-1070, 2009.
%  [3] Marcel van Gerven, Ali Bahramisharif, Tom Heskes and Ole Jensen, "Selecting features for BCI control based on a covert spatial attention paradigm."
%      Neural Networks 22 (9), 1271-1277, 2009
%  [4] Gehring, W.J., Coles, M.G.H., Meyer, D.E., Donchin, E.
%      "The error-related negativity: an event-related brain potential accompanying errors." 
%      Psychophysiology 27, 34-41, 1990
%  [5] Oliveira, F.T.P., McDonald, J.J., Goodman, D.
%      "Performance monitoring in the anterior cingulate is not all error related: expectancy deviation and the representation of action-outcome associations"
%      Journal of Cognitive Neuroscience. 19(12), 1994-2004, 2007
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-06-25

result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'lowpass',[2 14], 'epoch',[-1.5 1.5], 'srate',50, ...
    ... % learning defaults:
    'learner','dal', 'normalizers',{-0.25 -0.25}, ...
    'featureadapt',@scp_getscale,'featureextract',@scp_doscale));

% compute temporal & spatial normalizers
function mdl = scp_getscale(data,opts)
X = num2cell(data.data,[1 2]);
mdl.P = {cov_shrink(cat(2,X{:})')^opts.normalizers{1},cov_shrink(cat(1,X{:}))^opts.normalizers{2}};
mdl.vectorize = ~any(strcmp(opts.learner,'dal'));

% normalize and optionally vectorize the data
function X = scp_doscale(data,mdl)
X = data.data;
for t=1:size(X,3)
	X(:,:,t) = mdl.P{1}*X(:,:,t)*mdl.P{2}; end
if mdl.vectorize
    X = double(reshape(X,[],size(X,3))'); end
