function result = para_multiband_csp(data, mode, varargin)
% Result = para_multiband_csp(Input-Data, Operation-Mode, Options...)
% Paradigm for complex oscillatory processes using the multi-band Common Spatial Patterns (CSP) algorithm.
%
% Multi-band CSP [1,2] is a simple extension of the basic CSP method (see para_csp), in which for each of several time/frequency filters a set of CSP filters is learned,
% followed by log-variance feature extraction, concatenation of all features (over all chosen spectral filters) and subsequent machine learning. It is not 
% a general replacement for CSP due to the problem of overfitting, but is very useful whenever oscillatory processes in different frequency bands (and with
% different spatial topographies) are jointly active, and their concerted behavior must be taken into account for a given prediction task. Multiband-CSP can also
% be used to capture oscillations in multiple time windows, instead of frequency windows (for example for the detection of complex event-related dynamics).
%
% Since the dimensionality of the feature space is larger than in CSP, and since complex interactions may be present, a more complex classifier than the default
% LDA may be necessary to learn an appropriate model. On the other hand, more flexibility amplifies the risk of overfitting (especially with only little 
% calibration data), so that the performance should always be compared to standard CSP (and Spec-CSP). Another reason is that complex (relevant) interactions 
% between different frequency bands are seemingly rarely observed in practice. The most important user-configurable parameters are the selection regions in time 
% and frequency and the learner component. 
%
% Typical applications would be those in which either complex event-related oscillatory dynamics happen (for example when reacting to a particular
% stimulus) and/or where non-trivial interactions between frequency bands (e.g. alpha/theta) are relevant, such as, for example, in workload measurements.
% 
% Example: Consider a calibration dataset in which a subject is maintaining and updating different number of items in his/her working memory at different
% times, e.g. while performing the n-Back task [2]. Events with types 'n1','n2','n3' indicate challenge stimuli in which the respective number of items is 
% being processed by the person. The goal is to be able to predict the working-memory load of the person following the presentation of such a memory-related 
% challenge. An epoch of 3 seconds relative to each challenge is selected, and three different regions are chosen, two of them over the entire interval, covering
% the theta and alpha ryhthm, respectively, and one region that is restricted to a Gaussian-shaped window around the time of heaviest cognitive processing.
% The three regions are specified as a cell array of flt_select parameters.
%
%   dataset = io_loadset('datasets/mary/nback.eeg')
%   [loss,model,stats] = bci_train({'data',dataset, 'paradigm',@para_multiband_csp}, 'events',{'n1','n2','n3'}, ...
%       'epoch',[-0.5 2.5], 'regions',{{'freq' [4 6]},{'freq' [7 15]},{'time' {'gauss',[0.25 1.25]} 'freq' [7 15]}}, 'learner','logreg')
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'srate':   sampling rate, in Hz (default: 200)
%                     * 'window':  epoch window, in seconds relative to event (default: [0.5 3.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'regions':  cell array of time/frequency regions, each a cell array of name-value pairs as allowed by flt_select
%                                   (default: {{'freq' [7 15]},{'freq' [15 30]}})
%                     * 'patterns': the number of pattern pairs that CSP should produce; feature-space dimension will be twice this number (default: 3)
%                     * 'learner': learning function to be used on the features; (default: 'lda')
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References;
%   [1] Quadrianto Novi, Cuntai Guan, Tran Huy Dat, and Ping Xue, "Sub-band Common Spatial Pattern (SBCSP) for Brain-Computer Interface"
%       Proceedings of the 3rd International IEEE EMBS Conference on Neural Engineering Kohala Coast, Hawaii, USA, May 2-5, 2007
%   [2] Kai K. Ang, Zhang Y. Chin, Haihong Zhang, Cuntai Guan, "Filter Bank Common Spatial Pattern (FBCSP) in Brain-Computer Interface"
%       In 2008 IEEE International Joint Conference on Neural Networks (IEEE World Congress on Computational Intelligence) (June 2008), pp. 2390-2397.
%   [3] Owen, A. M., McMillan, K. M., Laird,A. R. & Bullmore, E. "N-back working memory paradigm: A meta-analysis of normative functional neuroimaging studies."
%       Human Brain Mapping, 25, 46-59, 2005
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-29


result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'regions', {{'freq' [7 15]},{'freq' [15 30]}}, 'epoch',[0.5 3.5], 'srate',200, ...
    ... % learning defaults:
    'patterns',3, 'learner','lda', ...
    'featureadapt',@csp_train,'featureextract',@csp_predict),'binary','featureadapt');


% obtain CSP spatial filters
function mdl = csp_train(data,opts)
filters = [];
if ~iscell(opts.regions{1})
    opts.regions = {opts.regions}; end
for region = 1:length(opts.regions)
    for k=1:2
        trials{k} = exp_eval(flt_select(set_picktrials(data,'rank',k),opts.regions{region}{:}));
        covar{k} = cov(reshape(trials{k}.data,size(trials{k}.data,1),[])');
        covar{k}(~isfinite(covar{k})) = 0;
    end
    [V,D] = eig(covar{1},covar{1}+covar{2}); %#ok<NASGU>
    filters = [filters V(:,[1:opts.patterns end-opts.patterns+1:end])];
end
mdl = struct('opts',{opts},'filters',{filters});
    
% extract filtered log-variance features
function X = csp_predict(data,mdl)
R = length(mdl.opts.regions);
F = size(mdl.filters,2);
T = size(data.data,3);
X = zeros(T,F);
for region = 1:R
    fltdata = exp_eval(flt_select(data,mdl.opts.regions{region}{:}));
    inds = (region-1)*(F/R)+(1:(F/R));
    for t=1:T
        X(t,inds) = log(var(fltdata.data(:,:,t)' * mdl.filters(:,inds))); end
end