function result = para_skeleton(data, mode, varargin)
% Result = para_skeleton(Input-Data, Operation-Mode, Options...)
% Generic Signal Processing -> Feature Extraction -> Machine Learning BCI skeleton; can be used to derive specialized paradigms.
%
% Most BCI paradigms are implemented as a sequence of three major stages: Signal Processing, Feature Extraction and Machine Learning (see also bci_train).
% The Signal Processing stage operates on time series (plus meta-data), represented as EEGLAB datasets, and may contain multiple sub-stages, which together
% form a filter graph. The data passed from node to node in this graph is either continuous or epoched EEGLAB datasets, and may contain rich annotations, such as 
% channel locations, ICA decompositions, DIPFIT models, etc. The nodes are largely general-purpose filter components which are shared among many paradigms,
% and most of them are found in filters/flt_* and dataset_ops/set_*. Furthermore, for almost all paradigms, a default order of these stages can be defined,
% because several filters can be arbitrarily ordered w.r.t. each other (linear operators), and other filters give reasonable results only when executed before
% or after certain critical other stages (for example, transform into variance or power spectrum should always be preceded by spatial filtering).
% Finally, there are efficiency considerations, such as running heavy processing preferably after dimensionality reduction (such as channel selection 
% and resampling). The default Signal Processing pipeline of the toolbox is a meta-filter which implements this default ordering (and in which any step
% can be conveniently enabled or disabled by the user).
%
% The simplest filter components are stateless (such as a surface laplacian filter) and operate on each sample individually, while other filters are 
% stateful (and have a certain "memory" of past data), such as FIR and IIR filters. Some of the stateful filters are time-variant (such as signal
% standardization) and some of those are adaptive (such as ICA). Some adaptive filters may be unsupervised, and others may depend on the target variable.
% The majority of filters is causal (i.e. does not need data from future samples to compute the transformed version of some sample) and can therefore 
% applied online, while some filters are non-causal (e.g., the signal envelope and zero-phase FIR filter), and can only be used for offline and/or 
% neuroscientific analyses. Finally, most filters operate on continuous data, while some filters operate on epoched/segmented data (such as, 
% for example, time window selection or fourier transform). All of these filter components are maintained in a unified framework.
%
% Following the Signal Processing stage, most paradigms implement the Feature Extraction stage (especially those paradigms which do not implement 
% adaptive statistical signal processing), in which signals are transformed into sets of feature vectors. At this stage, signal meta-data is 
% largely stripped off. The feature extraction performed by some paradigms is non-adaptive (such as windowed means of log-variance), while it is
% adaptive (and usually supervised) for others (e.g. CSP). Feature vectors can be viewed as (labeled) points in a high-dimensional space, the 
% feature space, which serves as the interface to the last stage, the machine learning.
%
% The machine learning stage is the last processing stage, and is implemented by most paradigms. Machine learning, being a sub-field of computer science,
% is a well-defined computational framework which lends itself well to be standardized. The machine learning stage is practically always adaptive, and 
% contains of a 'learning' step and a 'prediction' step. In the learning step, labeled sets of feature vectors are received, processed & analyzed, 
% their distribution w.r.t. the target variables (labels) is estimated, and a predictive model (or prediction function) incorporating these 
% relations is computed and emitted. In the prediction step, the previously computed predictive model is applied to individual feature vectors to 
% predict their label/target value (or a probability distribution over possible label/target values).
%
% Likewise, a paradigm has a 'learning' mode, in which data is received and a predictive model is computed (which is an arbitrary data structure that
% incorporates the state of all adaptive stages, e.g. feature extraction model and machine learning model), and a 'prediction' mode, in which a previously
% computed predictive model is used to map data to a prediction by sending it through all of the paradigm's stages. Finally, a paradigm has a 'preprocessing'
% mode, into which all the signal processing steps are factored. Separating the preprocessing from the other two stages gives more control to the framework
% (i.e. largely bci_train, onl_predict) in terms of the granularity (block size) of data that is fed through the processing stages, buffering, 
% transparent caching of intermediate results, transparent partitioning of datasets (for cross-validation, nested cross-validation and other resampling 
% techniques) and various optimizations (such as common subexpression elimination to minimize redundant computation and lazy evaluation).
% 
% The function para_skeleton represents a small sub-framework for the convenient implementation of paradigms that adhere to this overall system. 
% Paradigms may implement their functionality by calling into para_skeleton, setting some of its parameters in order to customize its standard system.
% Therefore, para_skeleton exposes a set of named parameters for each of the three stages. For signal processing, it exposes all the parameters of 
% flt_default, the default signal processing pipeline, allowing paradims to enable various pipeline stages without having to care about their relative
% order or efficient execution. For feature extraction, it exposes the 'featureextract' and 'featureadapt' parameters, which are function handles which
% implement the feature extraction and feature adaption (if any) step of this processing phase. Very few constraints are imposed on the type of 
% inputs, outputs and internal processing of these functions. For the machine learning stage, it exposes the 'learner' parameter, allowing to specify
% one of the ml_train*** / ml_predict*** functions for learning and prediction, respectively (only the *** tag needs to be specified, not the entire function).
%
% In addition to these options, para_skeleton offers some additional convenience facilities, e.g., the ability to use voting to extend two-class feature 
% extractors (such as Spec-CSP) to the general multi-class case, and assigns a minimal set of defaults for its parameters.
% 
% Paradigms making use of para_skeleton typically pass all user-specified parameters down to para_skeleton (so that the user has maximum control with no
% interference from the paradigm), and set up their own characteristic parameterization of para_skeleton as defaults for these user parameters.
%
% Para_skeleton contains no restrictions w.r.t. the type of data that is managed by it, and can be fed with EEGLAB datasets, as well as STUDY datasets,
% or any other custom dataset structure (aside from its own default feature extractor, which is, however, almost always overwritten) to facilitate the
% implementation of across-session / across-subject and across-task BCI paradigms like it facilitates the development of conventional BCI paradigms.
%
%
% In:
%       Input-Set       : some data set
%
%       Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                         'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                         'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%       User Options... : optional parameters of the paradigm; name-value pairs;
%                         --- for preprocessing
%                         * all parameters of flt_default are supported for preprocessing; the skeleton does not impose any defaults by itself
%                         --- for learning
%                         * 'featureadapt': the adaption stage of the feature extraction; function_handle, receives the preprocessed data,
%                                           an options struct (with all user options), and returns a model (which may just re-represent options)
%                                           (default: returns a struct with the specified options as the model)
%                         * 'featureextract': the feature extraction stage; function_handle, receives the preprocessed data and 
%                                             the model from the featureadapt stage, and returns a NxF array of N feature vectors (for F features)
%                                             (default: vectorizes each data epoch)
%                         * 'learner': is the machine learning function to be used on the features,
%                                      for example 'lda' (default: 'lda')
%                         * 'binary': whether some of the supplied functions support binary labels only,
%                                     can be either 'featureadapt' (meaning that the feature adaption is binary only) or [] (default: [])
%                         --- for prediction
%                         * 'model': the detector model (as produced during the learning)
%
% Out:
%       Result : depends on the Operation-Mode; 
%                * for 'preprocessing', this is the preprocessed dataset
%                * for 'learning', this is the learned model
%                * for 'prediction', this is the predictions produced by the model, given the data
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-04-29

% process input arguments & set defaults
opts = hlp_varargin2struct(varargin, ...
        'featureadapt',@(data,opts) opts, ...
        'featureextract',@(data,model) squeeze(reshape(data.data,[],1,size(data.data,3)))', ...
        'learner','lda', ....
        'binary',[]);

switch mode
    case 'preprocessing'
        % apply default signal processing
        result = flt_default(data, opts);
    case 'learning'
        classes = unique(set_gettarget(data));
        if ~(strcmp(opts.binary,'featureadapt') && numel(classes) > 2)
            % learn a model
            [result.featuremodel,result.predictivemodel] = learn_model(data,opts);
        else
            % binary stage and more than two classes: learn 1-vs-1 models for voting
            result.classes = classes;
            for i=1:length(classes)
                for j=i+1:length(classes)
                    [result.voting{i,j}.featuremodel,result.voting{i,j}.predictivemodel] = learn_model(exp_eval(set_picktrials(data,'rank',{i,j})),opts); end
            end
        end
        result.opts = opts;
    case 'prediction'
        if ~isfield(opts.model,'voting')
            % predict given the extracted features and the model
            result = ml_predict(opts.model.opts.featureextract(data,opts.model.featuremodel), opts.model.predictivemodel);
        else
            % 1-vs-1 voting is necessary, construct the aggregate result
            trialcount = exp_eval(set_partition(data,[]));
            result = {'disc' , zeros(trialcount,length(opts.model.classes)), opts.model.classes};
            % vote, adding up the probabilities from each vote
            for i=1:length(opts.model.classes)
                for j=i+1:length(opts.model.classes)
                    outcome = ml_predict(opts.model.opts.featureextract(data,opts.model.voting{i,j}.featuremodel), opts.model.voting{i,j}.predictivemodel);
                    result{2}(:,[i j]) = result{2}(:,[i j]) + outcome{2};
                end
            end
            % renormalize probabilities
            result{2} = result{2} ./ repmat(sum(result{2},2),1,size(result{2},2));
        end
end


function [featuremodel,predictivemodel] = learn_model(data,opts)
% adapt the feature extractor
featuremodel = opts.featureadapt(data,opts);
% extract features & learn a predictive model
predictivemodel = ml_train({opts.featureextract(data,featuremodel),set_gettarget(data)}, opts.learner);
