function result = para_speccsp(data, mode, varargin)
% Result = para_speccsp(Input-Data, Operation-Mode, Options...)
% Paradigm for oscillatory processes using the Spectrally weighted Common Spatial Patterns (Spec-CSP) algorithm.
%
% The Spec-CSP paradigm [1] is a more advanced variant of CSP, developed for the Berlin Brain-Computer Interface (BBCI); the primary focus was motor imagery BCI, 
% but the algorithm was designed from the outset to be applicable for a wider range of applications. The implementation closely follows the TR [2].
%
% The paradigm is applicable to the majority of oscillatory processes, and is the most advanced spatio-spectrally adaptive method that is currently 
% provided in the toolbox. Whenever the exact frequency and location of some (conjectured) oscillatory process is not known exactly, Spec-CSP can be used,
% and typically gives better results than CSP with an appropriately unrestricted (e.g., broad-band) spectral filter. Several other methods exist to adapt
% the spectrum to a process of interest, among others Common Spatio-Spectral Patterns [3], Common Sparse Spectral Spatial Pattern [4], r^2-based heuristics [5],
% automated parameter search, and manual selection based on visual inspection. Several of these methods have been shown to give approx. comparable results [2].
% An alternative and competitive method, especially when there complex interactions between frequency bands and time periods are to be modeled is the Dual-Augmented
% Lagrange paradigm (para_dal/para_dal_hf).
%
% The method iteratively optimizes spatial and spectral filters in alternation and extracts log-variance features from the resulting (filtered) data. 
% These features are subsequently processed by a (typically simple) machine learning component, by default LDA. Learning is therefore significantly slower 
% than CSP. An option is to impose custom prior knowledge on the relevant data spectrum, for example by placing a focus on the alpha rhythm, without ruling
% out other frequencies. Note that there are parameters which constrain the spectrum: one is the frequency prior and the other is the spectral filter that 
% is applied before running the alorithm; both need to be adapted when the considered spectrum shall be extended (e.g. to high-gamma oscillations). Other 
% parameters which are frequently adapted are the time window of interest and the learner component (e.g., logistic regression is a good alternative choice).
% 
% Some application areas include detection of major brain rhythm modulations (e.g. theta, alpha, beta, gamma), for example related to relaxation/stress, 
% aspects of workload, emotion, sensori-motor imagery, and in general cortical idle oscillations in various modalities.
%
% Example: Consider the goal of predicting the emotion felt by a person at a given time. A possible calibration dataset for this task would contain a sequence
% of blocks in each of which the subject is one out of several possible emotions, indicated by events 'e1','e2','e3','e4' covering these blocks in regular 
% intervals, for example induced via guided imagery [6].
%  
%   calib = io_loadset('datasets/bruce/emotions.eeg')
%   [loss,model,stats] = bci_train({'data',calib,'paradigm',@para_speccsp}, 'epoch',[-2.5 2.5], 'events',{'e1','e2','e3','e4'});
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'spectrum':  spectral filter (default: [5 45])
%                     * 'srate':   sampling rate, in Hz (default: 200)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [0.5 3.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'patterns': the number of pattern pairs that Spec-CSP should produce; feature-space dimension will be twice this number (default: 3)
%                     * 'prior': function of frequency implementing a prior (default: @(f) f>=7 & f<=30)
%                     * 'pp': regularization parameter p' (default: 0), can be optimized using search(-1:0.5:1)
%                     * 'qp': regularization parameter q' (default: 1), can be optimized using search(0:0.5:4)
%                     * 'steps': number of iterations (spatial optimization, followed by spectral optimization) to execute (default: 3)
%                     * 'learner': learning function to be used on the features; (default: 'lda')
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References:
%  [1] Tomioka, R., Dornhege, G., Aihara, K., and Mueller, K.-R. "An iterative algorithm for spatio-temporal filter optimization." 
%      In Proceedings of the 3rd International Brain-Computer Interface Workshop and Training Course 2006.
%  [2] Ryota Tomioka, Guido Dornhege, Guido Nolte, Benjamin Blankertz, Kazuyuki Aihara, and Klaus-Robert Mueller 
%      "Spectrally Weighted Common Spatial Pattern Algorithm for Single Trial EEG Classification",
%      Mathematical Engineering Technical Reports (METR-2006-40), July 2006.
%  [3] Steven Lemm, Benjamin Blankertz, Gabriel Curio, and Klaus-Robert Mller. 
%      "Spatio-spectral filters for improving classification of single trial EEG." 
%      IEEE Trans Biomed Eng, 52(9):1541-1548, 2005.
%  [4] G. Dornhege, B. Blankertz, M. Krauledat, F. Losch, G. Curio, and K.-R. Mller, 
%      "Combined optimization of spatial and temporal filters for improving brain-computer interfacing,"
%      IEEE Transactions on Biomedical Engineering, vol. 53, no. 11, pp. 2274?2281, 2006.
%  [5] Benjamin Blankertz, Ryota Tomioka, Steven Lemm, Motoaki Kawanabe, and Klaus-Robert Mueller. 
%      "Optimizing spatial filters for robust EEG single-trial analysis." 
%      IEEE Signal Process Mag, 25(1):41-56, January 2008
%  [6] Onton J & Makeig S. "Broadband high-frequency EEG dynamics during emotion imagination."
%      Frontiers in Human Neuroscience, 2009. 
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-29

result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'spectrum',[5 45], 'epoch',[0.5 3.5], 'srate',200, ...
    ... % learning defaults:
    'patterns',3, 'pp',0, 'qp',1, 'prior',@(f)f>=7&f<=30, 'steps',3, 'learner','lda', ...
    'featureadapt',@speccsp_adaptfeatures,'featureextract',@speccsp_extractfeatures,'binary','featureadapt'));

% compute the Spec-CSP feature space
function model = speccsp_adaptfeatures(data,opts)
% read a few parameters from the options (and re-parameterize the hyper-parameters p' and q' into p and q)
[n_of,p,q] = deal(opts.patterns,opts.pp+opts.qp,opts.qp);
% number of C=Channels, S=Samples and T=Trials #ok<NASGU>
[C,S,dum] = size(data.data); %#ok<NASGU>
% build a frequency table (one per DFT bin)
freqs = (0:S-1)*data.srate/S;
% evaluate the prior I
I = opts.prior(freqs); 
% and find table indices that are supported by the prior
bands = find(I);

% preprocessing
for c=1:2
    % compute the per-class epoched data X and its Fourier transform (along time), Xfft
    X{c} = exp_eval(set_picktrials(data,'rank',c));
    [C,S,T] = size(X{c}.data);
    Xfft{c} = fft(X{c}.data,[],2);
    % the full spectrum F of covariance matrices per every DFT bin and trial of the data
	F{c} = single(zeros(C,C,max(bands),T));
	for k=bands
		for t=1:T
			F{c}(:,:,k,t) = 2*real(Xfft{c}(:,k,t)*Xfft{c}(:,k,t)'); end
    end
    % compute the cross-spectrum V as an average over trials
    V{c} = mean(F{c},4);
end

% 1. initialize the filter set alpha and the number of filters J
J = 1; alpha{J}(bands) = 1;
% 2. for each step
for step=1:opts.steps
    % 3. for each set of spectral coefficients alpha{j} (j=1,...,J)
    for j=1:J
        % 4. calculate sensor covariance matrices for each class from alpha{j}
        for c = 1:2
            Sigma{c} = zeros(C);
            for b=bands
                Sigma{c} = Sigma{c} + alpha{j}(b)*V{c}(:,:,b); end
        end
        % 5. solve the generalized eigenvalue problem Eq. (2)
       [VV,DD] = eig(Sigma{1},Sigma{1}+Sigma{2});
       % and retain n_of top eigenvectors at both ends of the eigenvalue spectrum...
       W{j} = {VV(:,1:n_of), VV(:,end-n_of+1:end)};
       % as well as the top eigenvalue for each class
       lambda(j,:) = [DD(1), DD(end)]; 
    end
    % 7. set W{c} from all W{j}{c} such that lambda(j,c) is minimal/maximal over j
    W = {W{argmin(lambda(:,1))}{1}, W{argmax(lambda(:,2))}{2}};
    % 8. for each projection w in the concatenated [W{1},W{2}]...
    Wcat = [W{1} W{2}]; J = 2*n_of;
    for j=1:J
        w = Wcat(:,j);
        % 9. calcualate (across trials within each class) mean and variance of the w-projected cross-spectrum components 
        for c=1:2
            % part of Eq. (3)
            s{c} = zeros(size(F{c},4),max(bands));
            for k=bands
                for t = 1:size(s{c},1)
                    s{c}(t,k) = w'*F{c}(:,:,k,t)*w; end
            end
            mu_s{c} = mean(s{c});
            var_s{c} = var(s{c});            
        end
        % 10. update alpha{j} according to Eqs. (4) and (5)
        for c=1:2
            for k=bands
                % Eq. (4)
                alpha_opt{c}(k) = max(0, (mu_s{c}(k)-mu_s{3-c}(k)) / (var_s{1}(k) + var_s{2}(k)) );
                % Eq. (5), with prior from Eq. (6)
                alpha_tmp{c}(k) = alpha_opt{c}(k).^q * (I(k) * (mu_s{1}(k) + mu_s{2}(k))/2).^p;
            end            
        end
        % ... as the maximum for both classes
        alpha{j} = max(alpha_tmp{1},alpha_tmp{2});        
        % and normalize alpha{j} so that it sums to unity
        alpha{j} = alpha{j} / sum(alpha{j});
    end
end
alpha = [vertcat(alpha{:})'; zeros(S-length(alpha{1}),length(alpha))];
model = struct('W',{Wcat},'alpha',{alpha},'freqs',{freqs},'bands',{bands});

% extract filtered log-variance features
function X = speccsp_extractfeatures(data,model)
X = zeros(size(data,3),size(model.W,2));
for t=1:size(data.data,3)
    X(t,:) = log(var(2*real(ifft(model.alpha.*fft(data.data(:,:,t)'*model.W))))); end
