function [restable,xdims,ydims,patterns,accs,colmap] = vis_hyperspectrum(eeg,varargin)
% [Table,Time-Lim,Freq-Lim,Patterns,Accs,Colmap] = vis_hyperspectrum(EEG, Options...)
% calculate a CSP-based spectral discriminability plot
%
% In:
%       EEG     : epoched EEG set
%
%       Options : name-value pairs; see code for help & defaults
%
% Out:
%       Table    : the data table, containing accuracy estimates; Frequency x Time
%       Time-Lim : time limits of the table [lo,hi]
%       Freq-Lim : frequency limits of the table [lo,hi]
%       Patterns : CSP patterns table, contains an array of the best CSP patterns per T/F, fold, and class
%                  dims: {TimeSlices x FreqSlices x Folds x Classes} x Channels
%       Accs     : accuracies for the patterns {TimeSlices x FreqSlices}
%       Colmap   : color map containing color-embedded CSP patterns, if embedding enabled
%
% Explanation:
%       Plots the expected CSP+LDA classification accuracy for each time and frequency in the epoch window.
%       The expected accuracy underestimates the accuracy that can be gained by using large time windows and filters in a real CSP application.
%
%                               Christian Kothe, FG MMS, TU-Berlin

hann = @(N) 0.5*(1-cos(2*pi*(0:N-1)/(N-1)))';

% read the options & assign defaults
opts = hlp_varargin2struct(varargin,...    
            ... % windowing control
            'window', hann(150), ...        % sliding window type; larger means less temporal and more spectral resolution (and vice versa)
            'overlap', 135, ...             % number of overlapping samples between windowed estimations (the closer this is is to length(window), the slower it gets
            ... % scoring control
            'classes', {1,2}, ...           % default class selection for comparison
            'foldness', 10, ...             % chrono CV foldness
            'margin', 5, ...                % chrono CV safety margin, in trials
            'scoring', 'pragmatic', ...     % the scoring type: 'pragmatic' -> mean(nof=1,nof=3); 'optimistic' -> max(nof=1,nof=3)
            'bandrange', -2:2, ...          % micro band range around each measured band, for the CSP computation (in FFT bins)
            'scorefilter', [1 1 1]/3, ...   % scoring post-processing filter; empty means unfiltered
            'priorflt', [5 40], ...         % prior filter; restricts the printed bands
            'classifier', 'lda', ...        % e.g. 'lda','qda','svm','logreg' (or {trainfun,testfun}, e.g. {@ml_trainlda,@ml_testlda})
            ... % display control
            'preview', 1, ...               % show a live preview of the image being created
            'colormap', 'jet', ...          % the color map to be used for display
            'caxis', [0.45 0.75], ...       % color axis scaling; can be a [lo hi] array, or 'auto'
            ... % optional color-space embedding
            'embed', 0, ...                 % whether to embed the patterns into color space... (SLOW!)
            'precision', 0.5, ...           % overall sampling precision (using a value < 1 results in a sparser subset of all data being used)
            'perplexity', 20, ...           % a quantity related to the entropy in the dataset (more if the dataset is larger and more complicated...)
            'chance_level', 0.5, ...        % chance levels is mapped to black; default: chance level in the data set
            ... % saving
            'position', [0 0 1200 480], ... % figure size
            'plotfile', 'hyperspec_*.pdf'...% file name pattern
            );

warning off MATLAB:log:logOfZero
warning off MATLAB:illConditionedMatrix
warning off MATLAB:nearlySingularMatrix
warning off MATLAB:singularMatrix

% support specification of an optional CSP-specific logreg classifier
if ischar(opts.classifier) && strcmp(opts.classifier, 'logreg-csp')
    load csp_priors; opts.classifier = {@(set)ml_trainlogreg(set,priors_full),@ml_predictlogreg}; end
% support classifier specification as a string
if ischar(opts.classifier)
    opts.classifier = {eval(['@ml_train' opts.classifier]), eval(['@ml_predict' opts.classifier])}; end
% support classifier specification as cell array
if iscell(opts.classifier)
    opts.classifier = struct('train',opts.classifier{1}, 'test', opts.classifier{2}); end

% extract the per-class epoch data and a few more eeg parameters
srate = eeg.srate;
[C,S,T] = size(eeg.data);
for c=1:2
    epochs{c} = set_picktrials(eeg,'rank',opts.classes{c}); epochs{c} = epochs{c}.data; end

% replicate the window function for efficient application
repwnd = repmat(opts.window',[C 1]);

% create crossval index sets
for c=[1 2]
    trialcnt = size(epochs{c},3);
    % for each fold...
    for f=0:opts.foldness-1
        % choose the indices for the test set
        testinds = 1+floor(f*trialcnt/opts.foldness):min(trialcnt,floor((f+1)*trialcnt/opts.foldness));
        % choose the rest for training
        allinds = 1:trialcnt;
        % but also reserve a safety margin around the test indices
        for j=0:opts.margin
            allinds([max(1,testinds-j) min(trialcnt,testinds+j)]) = 0; end
        testids{c,f+1} = testinds;
        trainids{c,f+1} = find(allinds);
    end
end

% get window parameters
wndlen = length(opts.window);
offsets = 1:(wndlen-opts.overlap):(S-wndlen);
scoretable = zeros(ceil((wndlen+1)/2), length(offsets));
% and infer the FFT frequency bands
freqs = [0:ceil((wndlen+1)/2)-1]*srate/wndlen;
bands = find((freqs >= opts.priorflt(1)) & (freqs <= opts.priorflt(2)));
% determine plot axes & coloring
ydims = [freqs(bands(1)),freqs(bands(end))];
xdims = [eeg.xmin+wndlen/(2*eeg.srate),eeg.xmax-wndlen/(2*eeg.srate)];

% for each sub-window...
fprintf('this may take a while');
if ~isempty(opts.position)
    figure('Position',opts.position); end
colormap(opts.colormap); 

patterns = cell(length(offsets),max(bands),opts.foldness,2);

for k=1:length(offsets)
    wndofs = offsets(k);

    % calc the spectra
    for c = [1 2]
        % get N=#channes, S=#samples, T=#trials
        [N,x,T] = size(epochs{c}); S=wndlen;
        preproctrials{c} = zeros(N,S,T);
        % for each trial
        for i=1:T
            % extract the relevant data portion
            E = squeeze(epochs{c}(:,wndofs:wndofs+wndlen-1,i)).*repwnd;
            % subtract the mean for each channel
            preproctrials{c}(:,1:size(E,2),i) = E-repmat(mean(E,2),[1 wndlen]);
        end
        % calc the spectral density of the data
        ffttrials{c} = fft(preproctrials{c},[],2);
        % calc the cross-spectra
        uber_spec{c} = uber_spectrum(ffttrials{c},bands);
        [cross_spec{c} oldids{c}] = rebuild_cross_spectrum(uber_spec{c},[],[],1:T,bands);
    end

    % --- crossval score computation ---

    cvscore=[]; cspec = cross_spec;
    % for each fold...
    for f=1:opts.foldness
        % update the cross spectra
        for c=[1 2]
            [cspec{c} oldids{c}] = rebuild_cross_spectrum(uber_spec{c},cspec{c},oldids{c},trainids{c,f},bands); end
        % compute CSP patterns for each band
        for band = bands
            [patts vars] = ml_corecsp({mean(cspec{1}(:,:,max(min(bands),min(band+opts.bandrange,max(bands)))),3), mean(cspec{2}(:,:,max(min(bands),min(band+opts.bandrange,max(bands)))),3)},3);
            % for nof's of 1 and 3
            cpatterns{1}(:,:,band) = patts([1 end],:);
            cpatterns{3}(:,:,band) = patts;
            patterns{k,band,f,1} = patts(1,:);
            patterns{k,band,f,2} = patts(end,:);
        end
        % generate features, for 2 and 6 dimensions
        for nofcnt = [1 3]
            for c=[1 2]
                % training
                cfeatures{c} = log(power_project(ffttrials{c}(:,:,trainids{c,f}),cpatterns{nofcnt},bands));
                % testing
                ctestfeats{c} = log(power_project(ffttrials{c}(:,:,testids{c,f}),cpatterns{nofcnt},bands));
            end
            % compute the accuracy
            for band = bands
                % train classifier
                model = opts.classifier.train({(squeeze(cfeatures{1}(:,band,:))),(squeeze(cfeatures{2}(:,band,:)))});
                % and classify the test data
                for c=[1 2]
                    labels{c} = hlp_maplabels(opts.classifier.test((squeeze(ctestfeats{c}(:,band,:))),model),'classindices');
                    trials_correct(c) = sum(labels{c} == c);
                end
                cvscore{nofcnt}(band,f) = (trials_correct(1)+trials_correct(2)) / (length(labels{1})+length(labels{2}));
            end
        end
    end

    % create the combined crossval score
    for i=[1 3] 
        cvscore{i}(isnan(cvscore{i})) = 0.5; % assume equal class priors
        cvscore{i} = mean(cvscore{i}'); 
    end
    
    if strcmp(opts.scoring,'pragmatic') 
        cvscore = mean([cvscore{1};cvscore{3}]);
    elseif strcmp(opts.scoring,'optimistic') 
        cvscore = max([cvscore{1};cvscore{3}]);
    else
        error('unsupported scoring mode'); 
    end

    % create a smoothed version of the cvscore
    if isempty(opts.scorefilter) 
        cscore = cvscore; 
    else
        cscore = filtfilt(opts.scorefilter,1,cvscore); 
    end

    accs(k,bands) = cscore(bands);
    scoretable(bands,k) = cscore(bands);
    % display the progress indicator
    if opts.preview
        draw_image(scoretable(bands,:),xdims,ydims,opts.caxis,patterns,eeg.chanlocs);
    else
        fprintf('.');        
    end
end
fprintf('\n');
restable = draw_image(scoretable(bands,:),xdims,ydims,opts.caxis,patterns,eeg.chanlocs);

% save figure...
if ~isempty(opts.plotfile)
    save2pdf(strrep(opts.plotfile,'*','acc'),gcf); end

if opts.embed
    disp('now coloring the patterns; this may take a while.');
    % grab parameters...
    acc = accs;    
    patts = patterns;
    precision = opts.precision;
    perplexity = opts.perplexity;
    if isempty(opts.chance_level)
        chance_level = sum(eeg.classidx==1)/length(eeg.classidx); 
        chance_level = max(chance_level, 1-chance_level);
    else
        chance_level = opts.chance_level;
    end

    % for each class...
    for p=1:2
        disp(['class ' num2str(p) ':']);
        disp(' sub-sampling patterns...');
        % adaptively sub-sample the patterns
        tmpacc = acc(:); tmpacc = tmpacc(find(tmpacc));
        minacc = mean(tmpacc)-2*std(tmpacc);
        maxacc = mean(tmpacc)+2*std(tmpacc);
        pattsubset = [];
        indices = [];
        for y=1:size(patts,2)
            for x=1:size(patts,1)
                cvfold = size(patts,3);
                % the sampling granularity is accuracy-dependent
                granularity = min(cvfold,floor(cvfold*(acc(x,y) - minacc)/(maxacc-minacc)));
                if granularity > 0
                    for k=1:cvfold/granularity:cvfold
                        pattsubset = [pattsubset; abs(patts{x,y,floor(k),p})];
                        indices(end+1,:) = [x,y];
                    end
                end
            end
        end

        % embed into 2 dimensions using t-SNE
        disp(' embedding patterns...');
        thesubset = pattsubset(ceil(1:1/precision:end),:);
        indices = indices(ceil(1:1/precision:end),:);
        [mapped,landmarks] = fast_tsne(thesubset, [], 2, 1);
        indices = indices(landmarks,:);

        % average over the folds
        disp(' averaging per-fold embedding...');
        sums = zeros(size(acc,1),size(acc,2),2);
        means = zeros(size(acc,1),size(acc,2),2);
        counts = zeros(size(acc,1),size(acc,2));
        for k=1:size(indices,1)
            sums(indices(k,1),indices(k,2),:) = squeeze(sums(indices(k,1),indices(k,2),:)) + mapped(k,:)';
            counts(indices(k,1),indices(k,2)) = counts(indices(k,1),indices(k,2)) + 1;
        end
        for x=1:size(acc,1)
            for y=1:size(acc,2)
                if counts(x,y) ~= 0
                    means(x,y,:) = sums(x,y,:) / counts(x,y);
                end
            end
        end

        % do a blur filtering
        kernel = ones(3)*1/9;
        means(:,:,1) = filter2(kernel,means(:,:,1));
        means(:,:,2) = filter2(kernel,means(:,:,2));

        % resample the patterns
        disp(' resampling patterns...');
        accb = acc(:,bands); means = means(:,bands,:);
        [u,v] = meshgrid(1:size(accb,2),1:size(accb,1));
        [ui,vi] = meshgrid(1:0.25:size(accb,2),1:0.25:size(accb,1));
        accb = interp2(u,v,accb,ui,vi);
        means2(:,:,1) = interp2(u,v,means(:,:,1),ui,vi);
        means2(:,:,2) = interp2(u,v,means(:,:,2),ui,vi);
        minx = min(min(means2(:,:,1))); maxx = max(max(means2(:,:,1)));
        miny = min(min(means2(:,:,2))); maxy = max(max(means2(:,:,2)));

        % embed into the color space
        disp(' mapping into color space...');
        colmap{p} = zeros([size(means2,2) size(means2,1) 3]);
        for x=1:size(means2,1)
            for y=1:size(means2,2)
                col = [(means2(x,y,1) - minx)/(maxx-minx),(means2(x,y,2) - miny)/(maxy-miny)];
                colmap{p}(y,x,:) = max(0,min(1,(accb(x,y)-chance_level)/(opts.caxis(2)-chance_level))) * ntsc2rgb([0.5, col(1)-0.5, col(2)-0.5]);
            end
        end
        % display
        global vis_data; vis_data = struct('xdims',{xdims},'ydims',{ydims},'patterns',{patterns},'chanlocs',{eeg.chanlocs});
        figure('Position',opts.position); image(colmap{p}, 'YData',ydims, 'XData',xdims,'ButtonDownFcn',@showpatterns);
        % save figure...
        if ~isempty(opts.plotfile)
            save2pdf(strrep(opts.plotfile,'*',['emb' num2str(p)]),gcf); end

    end
end


% draw the upsampled image
function restable = draw_image(table,xdims,ydims,cdims,patterns,chanlocs)
    global vis_data; vis_data = struct('xdims',{xdims},'ydims',{ydims},'patterns',{patterns},'chanlocs',{chanlocs});
    restable = table;
    [u,v] = meshgrid(1:size(restable,2),1:size(restable,1));
    [ui,vi] = meshgrid(1:0.25:size(restable,2),1:0.25:size(restable,1));
    restable = interp2(u,v,restable,ui,vi);    
    imagesc(restable,'YData',ydims, 'XData',xdims, 'ButtonDownFcn', @showpatterns);
    caxis(cdims);
    colorbar;
    drawnow;

% calculate the cross-bandpower for the given data set
% [Channels x Channels x Frequencies x Trials] = uber_spectrum([Channels x Frequencies x Trials], Bands)
function uber_spec = uber_spectrum(spec, bands)
	[C,S,T] = size(spec);
	uber_spec = zeros(C,C,max(bands),T);
	for f=bands
		for t=1:T
			uber_spec(:,:,f,t) = 2*real(spec(:,f,t)*spec(:,f,t)');
		end		
    end

    
% update the averaged cross-bandpower for each frequency and pair of channels, given the old & the new set of trials
% [Channels x Channels x Frequencies] = rebuild_cross_spectrum([Channels x Channels x Samples x Trials], [Channels x Channels x Frequencies], Trials, Trials, Bands)
function [newcspec newtrials] = rebuild_cross_spectrum(uspec,oldcspec,oldtrials,newtrials,bands)
    if isempty(oldcspec) || isempty(oldtrials)
        newcspec(:,:,bands) = mean(uspec(:,:,bands,newtrials),4);
    else
        added = setdiff(newtrials,oldtrials);
        removed = setdiff(oldtrials,newtrials);
        oldcspec(:,:,bands) = oldcspec(:,:,bands)*length(oldtrials);
        if isempty(added)            
            if isempty(removed)
                newcspec(:,:,bands) = oldcspec(:,:,bands);
            else
                newcspec(:,:,bands) = oldcspec(:,:,bands) - sum(uspec(:,:,bands,removed),4);
            end
        else
            if isempty(removed)
                newcspec(:,:,bands) = oldcspec(:,:,bands) + sum(uspec(:,:,bands,added),4);                
            else
                newcspec(:,:,bands) = oldcspec(:,:,bands) - sum(uspec(:,:,bands,removed),4) + sum(uspec(:,:,bands,added),4);    
            end
        end
        newcspec(:,:,bands) = newcspec(:,:,bands)*(1/length(newtrials));
    end    
    
    
% generate variance features for each frequency
% [Patterns x Frequencies x Trials] = power_project([Channels x Frequencies x Trials],[Patterns x Channels x Frequencies],Bands)
function spec_var = power_project(spec,patterns,bands)
	% get the data dimensions    
	[C,S,T] = size(spec);
    spec_var = zeros(size(patterns,1),ceil((S+1)/2),T);
    
	% for each projection
	for p=1:size(patterns,1)
        % replicate it
        V = repmat(squeeze(patterns(p,:,bands)),[1,1,T]);        
        % project trial spectra
        spec_var(p,bands,1:T) = abs(squeeze(sum(spec(:,bands,:).*V,1)));
    end


function showpatterns(x,y,z)
    % display patterns on mouse click
    global vis_data
    position = get(gca,'Currentpoint');
    xrel = (position(1,1)-vis_data.xdims(1))/(vis_data.xdims(2)-vis_data.xdims(1));
    yrel = (position(1,2)-vis_data.ydims(1))/(vis_data.ydims(2)-vis_data.ydims(1));
    xind = floor(size(vis_data.patterns,1)*xrel)+1;
    yind = floor(size(vis_data.patterns,2)*yrel)+1;
    pattern{1} = vis_data.patterns{xind,yind,1,1};
    pattern{2} = vis_data.patterns{xind,yind,1,2};
    figure; hold;
    subplot(1,2,1); vis_topo(pattern{1},vis_data.chanlocs);
    subplot(1,2,2); vis_topo(pattern{2},vis_data.chanlocs);
    title(['CSP patterns at ' num2str(position(1,1)) 's / ' num2str(position(1,2)) 'Hz.']);

    
% TODO:
% ... these could speed it up by x2:
% - get rid of repmat (gop_times)
% - get rid of squeeze? (maybe reorient matrices)
% - get rid of the dot product in LDA code (use *)
% - speed up the sum in the power_project (e.g. reorient spectrum)
