function figures = vis_scp(sets, channels, varargin)
% vis_scp(Input-Sets, Channels, Option-String, additional parameters...)
% visualize slow cortical potentials; advanced rewrite of the organically grown vis_erp, but not yet fully finished.
%
% In:
%      Input-Sets      :   the input eeg sets that should be compared
%                          can be a single set or a cell array of sets
%      Channels        :   the channels that should be displayed; either numeric or cell-string array
%                          one plot per channel is generated; the shape of this array specifies the plot arragement
%
%      Additional Parameters: see below
%
%                               Christian Kothe, FG MMS, TU-Berlin
%                               26.05.2008


opt = hlp_varargin2struct(varargin,  ...
                          ...
                          ... % --- preprocessing ---
                          ...
                          'prescaling',0.1, ...     % data pre-scaling
                          'map','_', ...            % arbitrary data mapping, e.g. _*0.1 or _2-_1
                          'mmh',0, ...              % do a miss-minus hit transform (requires exactly 2 classes)
                          ...
                          ... % --- content selection ---
                          ...
                          'means',1, ...            % plot means
                          'stds',1, ...             % plot standard deviations
                          'rsq',0, ...              % plot r-square values
                          'gauss',0, ...            % plot gaussian distributions
                          ...
                          ... % --- plot layout ---
                          ... % note: in descriptive strings, the substring <chan> is implicitly replaced by the Channel name
                          ...
                          'title', '<chan>: ERP', ...                    % plot title
                          'pdffile', 'data:/plots/channel_erps.pdf', ...  % pdf file to save the results (e.g. data:/myplot.pdf)
                          'labelfontsize', 10, ...  % font size of the plot & axis labels
                          'axisfontsize', 10, ...   % font size of the axis ticks
                          'titlefontsize', 10, ...  % font size of the title
                          'linewidth', 1, ...     % line with, for means etc.
                          'colors', {'r','g','b'},...           % plot colors, should be a cell-array
                          'xlabel', 'time (s)', ...  % x axis label
                          'ylabel', 'voltage (uV)', ...         % y axis label
                          'xtick', [], ...                      % x axis ticks (empty: as chosen by matlab)
                          'ytick', [], ...                      % y axis ticks (empty: as chosen by matlab)
                          'xlimit', [], ...                     % plot x limits
                          'ylimit', [], ...                     % plot y limits
                          'grid', 1);                           % grid

% --- data gathering ---

if iscell(sets)
    % handle vis_scp({l r}, ...);
    data = {};
    for c=1:length(sets)
        if isstruct(sets{c})
            data{c} = permute(sets{c}.data, [3 2 1]);
        else
            data{c} = permute(sets{c}, [3 2 1]);
        end
    end
else
    % handle vis_scp(epochs, ...);
    for c=1:max(sets.classidx)
        data{c} = set_picktrials(sets,'rank',c);
        data{c} = permute(data{c}.data, [3 2 1]);
    end
    
    % in this case, we are also able to remap the channel names to indices
    channels = set_chanid(sets, channels);
end

% --- data pre-processing ---

% scaling (according to scaling)
for d=1:length(data) data{d} = data{d}*opt.prescaling; end; 

% remapping (according to map)
mapping = lambda(opt.map);
if nargin(mapping) == 1
    % per-class map
    for d=1:length(data)
        data{d} = mapping(data{d});
    end;
else
    % cross-class map
    data = {mapping(data{:})};
end

% conversion to real numbers
for s=1:length(data)
    if ~isreal(data{s})
        disp('warning: replacing complex values by their absolute value');
        data{s} = abs(data{s});
    end
end

% miss-minus hit transformation
if opt.mmh
    if length(data) ~= 2
        error('exactly 2 classes are required for miss-minus hit plots');
    end
    data = {squeeze(mean(data{2}) - mean(data{1}))};
end


% --- graphics ---

% for each channel
for c = 1:numel(channels)
    % select the plot
    subplot(size(channels,1),size(channels,2),c);
    hold;

    % --- content generation ---

    % -- plot gaussian distributions --
    if opt.gauss
        % find image dimensions
        if opt.ylimit
            [min_data,max_data] = deal(opt.ylimit(1),opt.ylimit(2));
        else
            min_data = 1000000;
            max_data = -1000000;
            for d = 1:length(data)
                m = mean(data{d}(:,:,channels(c)));
                s = std(data{d}(:,:,channels(c)));
                min_data = min(min_data,min(m-s*3));
                max_data = max(max_data,max(m+s*3));
            end
        end
        % get one plane (r g b) for each set
        for d=1:length(data)
            img{d} = plane(data{d}(:,:,channels(c)), min_data, max_data);
        end
        % pad with zero-color channels
        for d=(length(data)+1):3
            img{d} = zeros(size(img{1},1), size(img{1},2));
        end            
        % we need to rescale the colors manually
        max_img = max([max(max(img{1})) max(max(img{2})) max(max(img{3}))]);
        img = cat(3, scale(img{1},0,max_img), scale(img{2},0,max_img), scale(img{3},0,max_img));
        image(img,'YData',[min_data max_data],'XData',[sets.xmin,sets.xmax]);
        ylim([min_data max_data]);
    end
    
    % -- plot means --
    if opt.means
        if ndims(data{1}) == 3
            % for each class
            for d = 1:length(data)
                m = mean(data{d}(:,:,channels(c)));
                plot(sets.times/1000,m,opt.colors{d});
            end
        else
            % for each class
            for d = 1:length(data)
                plot(sets.times/1000,data{d}(:,channels(c)),opt.colors{d});
            end            
        end
    end
    
    % -- plot standard deviations --
    if opt.stds && ndims(data{1})==3
        % for each class
        for d = 1:length(data)
            m = mean(data{d}(:,:,channels(c)));
            s = std(data{d}(:,:,channels(c)));
            plot(sets.times/1000,m-s,[opt.colors{d} '--']);
            plot(sets.times/1000,m+s,[opt.colors{d} '--']);
        end
    end
    
    % --- axis labeling ---
    try channame = sets.chanlocs(channels(c)).labels;        
    catch channame = num2str(channels(c)); end           
    xlabel(regexprep(opt.xlabel,'<chan>',channame),'FontSize',opt.labelfontsize);
    ylabel(regexprep(opt.ylabel,'<chan>',channame),'FontSize',opt.labelfontsize);
    if opt.title title(regexprep(opt.title,'<chan>',channame),'FontSize',opt.titlefontsize); end
    if opt.xlimit xlim(opt.xlimit); else xlim([sets.xmin, sets.xmax]); end
    if opt.ylimit ylim(opt.ylimit); end
    if opt.xtick set(gca,'xtick',opt.xtick); end
    if opt.ytick set(gca,'ytick',opt.ytick); end
    if opt.axisfontsize set(gca,'FontSize',opt.axisfontsize); end
    if opt.grid grid('on'); end
    
    hold;
end
 
try
   % save as pdf/png etc...
   save2pdf(opt.pdffile,gcf,600);
   %... and save as figure
   opt.pdffile(end-3:end) = '.fig';
   saveas(gcf,opt.pdffile);
catch,end



% calculate an image plane from a (trials x samples)-matrix
% display only the values between min_data and max_data
function [img mu sigma] = plane(mat, min_data, max_data)
    % get mu/sigma per time sample
    mu = mean(mat(:,:));
    sigma = std(mat(:,:));
    % generate the image
%    img = zeros(round(max_data-min_data), size(mat,2));
    img = zeros(1000, size(mat,2));    
    for x=1:size(img,1)
        img(x,:) = normpdf(((x/size(img,1))*(max_data-min_data)+min_data), mu, sigma);
    end

% normalize an array to 0..1 (min_mat and max_mat are optional)
function res = scale(mat, min_mat, max_mat)
    if (~exist('min_mat')) min_mat = min(min(mat)); end
    if (~exist('max_mat')) max_mat = max(max(mat)); end
    res = (mat - min_mat) * 1/(max_mat - min_mat);

% sample a normal probability distribution specified by mu/sigma at the time points given by x
function y = normpdf(x,mu,sigma)
    y = exp(-0.5 * ((x - mu)./sigma).^2) ./ (sqrt(2*pi) .* sigma);

    