function [minidx, inputs, outputs] = utl_gridsearch(func, varargin)
% [Min-Index, Inputs, Outputs] = utl_gridsearch(Function, Domain...)
% [Min-Index, Inputs, Outputs] = utl_gridsearch(Range-Format, Function, Domain...)
% exhaustive search over multiple possible arguments to a function (specified as arrays of possibilities), where all combinations
% are tried in the case of multiple argument arrays; the index of the minimum return value of the function is returned
%
% In:
%   Range-Format: argument range format, optional; either 'direct' (search ranges are directly specified as arrays),
%                 or 'clauses' (search ranges are specified using the search() clause). default: 'direct'
%
%   Function    : the function to optimize; should always return a real number (or NaN) as first output,
%                 taken as the objective value of the function; may have additional output arguments, which are also captured
%
%   Domain      : one-dimensional argument ranges to be searched; 
%                 * in 'direct' mode, each array determines the possible values at that place in the function's argument list, all combinations 
%                   of specified values are tried; for cell arrays, the search runs through all given cell contents at that position
%                 * in 'clauses' mode, search ranges for a given argument are specified using the search() expression in that place of the argument list
%
% Out:
%   Min-Index   : index of the minimum function return value (first return value if multiple ones are returned),
%                 out of all tried executions (see example); can be used as index into Inputs and/or Outputs
%
%   Inputs      : cell array of all tried input combinations (each a cell array of arguments)
%
%   Outputs     : cell array of all received function outputs, one for each input (each a cell array of outputs, since the
%                 function can have multiple outputs)
%
% Example:
%   For the three equivalent calls,
%     utl_gridsearch(@f, 1:2, {'a',{'b'},5:7,{5:7}}, [], {}, 'xy', {[1 2 3]});
%     utl_gridsearch('direct', @f, 1:2, {'a',{'b'},5:7,{5:7}}, [], {}, 'xy', {[1 2 3]});
%     utl_gridsearch('clauses', @f, search(1,2), search('a',{'b'},5:7,{5:7}), [], {}, search('x','y'), [1 2 3]);
%
%   ... the following 16 = 2x4x1x1x2x1 executions of f are compared:
%       f(1,'a',[],{},'x', [1 2 3])
%       f(1,'a',[],{},'y', [1 2 3])
%       f(1,{'b'},[],{},'x', [1 2 3])
%       f(1,{'b'},[],{},'y', [1 2 3])
%       f(1,5:7,[],{},'x', [1 2 3])
%       f(1,5:7,[],{},'y', [1 2 3])
%       f(1,{5:7},[],{},'x', [1 2 3])
%       f(1,{5:7},[],{},'y', [1 2 3])
%       f(2,'a',[],{},'x', [1 2 3])
%       f(2,'a',[],{},'y', [1 2 3])
%       f(2,{'b'},[],{},'x', [1 2 3])
%       f(2,{'b'},[],{},'y', [1 2 3])
%       f(2,5:7,[],{},'x', [1 2 3])
%       f(2,5:7,[],{},'y', [1 2 3])
%       f(2,{5:7},[],{},'x', [1 2 3])
%       f(2,{5:7},[],{},'y', [1 2 3])
%
% See also:
%   search()
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-06

% search-mode related processing
if ischar(func)
    [mode,func,varargin] = deal(func,varargin{1},varargin(2:end));
    if strcmp(mode,'clauses')
        for i=1:length(varargin)
            varargin{i} = hlp_flattensearch(varargin{i},'cell'); end
    elseif ~strcmp(mode,'direct')
        error('unsupported search mode.');
    end
end

% determine dimensions of the input grid
dims = cellfun(@length,varargin);
pitches = [1 cumprod(max(1,dims))];
combos = prod(max(1,dims));

args = varargin;
for i=1:length(args)
    % pre-unpack cell args for singleton arguments
    if iscell(args{i}) && dims(i) == 1
        args{i} = args{i}{1}; end
end
for c=1:combos
    for i=find(dims>1)
        % select current value for non-singleton arguments
        args{i} = varargin{i}(1+floor(mod((c-1)/pitches(i),dims(i))));
        % and unpack cell args
        if iscell(varargin{i})
            args{i} = args{i}{1}; end
    end
    outputs{c} = hlp_wrapresults(func,args{:});
    inputs{c} = args;
end

try
    minidx = hlp_getresult(2,@min,cellfun(@(x)x{1},outputs));
catch
    minidx = NaN;
end