function [action,result] = utl_memoize_lookup(exp,memo_ctrl)
% [Action,Result] = utl_memoize_lookup(Key-Expression,Memo-Locations)
% check for memoizability and/or availability of the given expression and decide on an action
% 
% In:
%  Key-Expression : an expression that uniquely identifies the object to be looked up;
%                   this expression is also required to be identical to the one stored in the looked-up object's .bcilab.expression field
% 
%  Memo-Locations : cell array of location specifiers, given as name-value pairs. possible names are currently 'disk' and 'memory', and the value
%                   is expected to be an expression that evaluates into 0 or 1, to indicate whether the given location is enabled for 
%                   storage/retrieval or not. The expression may use the symbol @expression to refer to the Key-Expression. The locations 
%                   are looked up in order, so the fastest locations should come first in the list.
%
% Out:
%   Action  : the action to be taken by the caller, one of:
%             'return' : result contains the object that shall be returned as first output argument
%             'memoize': result contains a cell array of store locations to be used for storing the result, when it is eventually computed
%                        - store locations begining with the fileep indicate disk locations, relative to the appropriate memoization domain
%                        - store locations beginning with a . indicate memory locations, relative to the in-memory cache data structure
%             'skip': do not memoize the result
% 
%   Result  : the payload to which the action applies (either the successfully looked up result or the memoization 
%             locations for a subsequent commit)
%
% Notes:
%   If a location further down in the list has the object in question but preceding ones do not, the object is implicitly commited 
%   (copied) to these locations.
%
%                                                     Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                                                     2010-05-23

global bcilab;

if ~exist('memo_ctrl','var') || (isempty(memo_ctrl) && ~iscell(memo_ctrl))
    memo_ctrl = hlp_lookup_symbol(@memoize);
    if ~isempty(memo_ctrl)
        memo_ctrl = utl_releasehold(memo_ctrl{1}); end
end

action = 'skip';
result = {};
% first check whether the current expression is matched by anything in memoize
if iscell(memo_ctrl)
    for i=1:2:length(memo_ctrl)
        % evaluate whether memoization is enabled right now
        enabled = exp_eval(hlp_microcache('lookup',@utl_replacerepeated,memo_ctrl{i+1}, {exp_rule(@expression,exp)}));
        if enabled
            % memoization is enabled for the following location
            location = memo_ctrl{i};
            exp_hash = hlp_microcache('hashes',@(x) num2str(145342 + hlp_fingerprint(x)),exp);            
            if strcmp(location,'disk')
                % DISK-BASED MEMOIZATION: compute the storage location according to some path schema
                memo_path = [filesep exp_hash(1:2) filesep exp_hash(3:end) '.mat'];
                % try all known disk paths for lookup
                for p = fieldnames(bcilab.cache.disk_paths)'
                    try
                        % try to load the file
                        location = bcilab.cache.disk_paths.(p{1});
                        file_location = [location.dir memo_path];
                        t0 = tic;
                        fprintf(['loading ' file_location '...']);
                        load(file_location,'obj');
                        fprintf(['complete.\n']);
                        
                        % delete the file if it is invalid for some reason (e.g. the machine crashed during saving)
                        if ~exist('obj','var') && strcmp('.mat',file_location(end-3:end))
                            delete(file_location); 
                            error('error');
                        end
                        % check whether the expression matches
                        if ~isstruct(obj) || ~isfield(obj, 'bcilab') || ~isfield(obj.bcilab,'expression')
                            warning('BCILAB:exp_begindef:memo_error','memoized object is not conformat; reverting...');
                            error('error');
                        end
                        if ~utl_same(obj.bcilab.expression,exp)
                            warning('BCILAB:exp_begindef:memo_error','exp_begindef: hash conflict between ');
                            disp(['* retrieved: ' exp_fullform(obj.bcilab.expression)]);
                            disp(['* requested: ' exp_fullform(exp)]);
                            error('error');
                        end
                        % got the correct result: make the function return it
                        if ~isempty(result) && strcmp(action,'memoize')
                            % ... but first commit it to the other higher-priority memory locations that don't yet have it
                            for r=1:length(result)
                                utl_memoize_commit(obj,result{r}); end
                        end                        

                        result = obj;                       
                        action = 'return';
                        try
                            % ... and record some statistics on the read speed of this location...
                            objinfo = whos('obj');
                            location.readstats(end+1) = struct('size',{objinfo.bytes},'time',{toc(t0)});
                            % write back the updated location record
                            bcilab.cache.disk_paths.(p{1}) = location;
                        catch,end
                        return;
                    catch,end
                end
                
                % value not yet stored: we want to memoize after the function's body has been evaluated, so remember the location data
                action = 'memoize';
                result{end+1} = memo_path;
            elseif strcmp(location,'memory')
                % MEMORY-BASED MEMOIZATION: compute the storage location according to a fieldname schema
                memo_field = ['x' exp_hash];
                % check if the requested object is present
                try
                    % found: return
                    obj = bcilab.cache.data.(memo_field);
                    % check whether the expression matches
                    if ~isstruct(obj) || ~isfield(obj, 'bcilab') || ~isfield(obj.bcilab,'expression')
                        warning('BCILAB:exp_begindef:memo_error','memoized object is not conformat; reverting...');
                        error('error');
                    end
                    if ~utl_same(obj.bcilab.expression,exp)
                        warning('BCILAB:exp_begindef:memo_error','exp_begindef: hash conflict between ');
                        disp(['* retrieved: ' exp_fullform(obj.bcilab.expression)]);
                        disp(['* requested: ' exp_fullform(exp)]);
                        error('error');
                    end                    
                    bcilab.cache.times.(memo_field) = cputime; % for the least-recently used policy
                    % got the correct result: make the function return it
                    if ~isempty(result) && strcmp(action,'memoize')
                        % ... but first commit it to the other higher-priority memory locations that don't yet have it
                        for r=1:length(result)
                            utl_memoize_commit(obj,result{r}); end
                    end
                    result = obj;
                    action = 'return';
                    return;
                catch,end
                % not found: remember to memoize later
                action = 'memoize';
                result{end+1} = ['.' memo_field];
            end
        end
    end
end
