function [model,stats] = utl_searchmodel(data, varargin)
% [Model,Stats] = utl_searchmodel(Data, Arguments...)
% search & return the best model, out of multiple possibilities (via cross-validation); specified via multiple possible training argument configurations
%
% In:
%   Data        :   some data that can be partitioned using index sets
%
%   Arguments   :   mandatory arguments:
%                   'trainer': training function; receives a partition of the data (as produced by the partitioner), 
%                              some further arguments (as specified in args), and returns a model (of any kind)
%
%                   'tester' : testing function; receives a model (as produced by the trainer) and a partition of the data
%                              (as produced by the partitioner), and returns a prediction for every index of the input data,
%                              in a format allowed by ml_predict
%
%                   'args': arguments to the training function, cell array (default: empty)
%                           specified as in utl_gridsearch (format controlled via argform)
%
%                   optional arguments (same as in utl_crossval, listed below for convenience)
%                   'scheme': cross-validation scheme, can be one of the following formats (default: 10)
%                     * 0: skip CV, return NaN and empty statistics
%                     * k: k-fold randomized CV (or, if 0<k<1, k-holdhout CV)
%                     * [r k]: r-time repartitioned k-fold randomized CV (or, if 0<k<1, k-holdout CV (with k a fraction))
%                     * 'loo': leave-one-out CV
%                     * {'chron', k} or {'block', k}: k-fold chronological/blockwise CV
%                     * {'chron', k, m} or {'block', k, m}: k-fold chronological/blockwise CV with m indices margin width (between training and test set)
%
%                   'partitioner': partitioning function for the data, receives three parameters: (data, index vector, packed trainer args OR model)
%                                   * if the index vector is empty, should return the highest index in the data
%                                   * otherwise, it should return data subindexed by the index vector
%                                  default: provides support for cell arrays, numeric arrays, struct arrays and {Data,Target} cell arrays
%                                  note: the third parameter is for convenience and may optionally be taken into account; trainer args are passed packed into 
%                                  a cell array for both index set generation and computation of the training partition(s), and the model is passed for the
%                                  computation of testing partitions
%
%                   'target': a function to derive the target variable from a partition of the data (as produced by the partitioner),
%                             for evaluation; the allowed format is anything that may be output by ml_predict
%                             default: provides support for {Data,Target} cell arrays
%
%                   'metric': metric to be employed, applied both in each fold and the aggregated data over all folds
%                              * function handle: a custom, user-supplied loss function; receives target data in the first 
%                                argument and prediction data in the second argument; each can be in any format that can be 
%                                produced by ml_predict (but can be expected to be mutually consistent).
%                                shall return a real number indicating the summary metric over all data, and optionally 
%                                additional statistics in a struct
%                              * string: use ml_calcloss, with 'metric' determining the loss type
%                              * default: use 'mcr','mse','nll','kld', depending on supplied target & prediction data formats
%
%                   'argform': format of the argument ranges, either 'direct' or 'clauses' (default: 'clauses')
%                               'direct': search ranges are directly specified as arrays
%                               'clauses': search ranges are specified using the search() clause
%
%                   'forcestats': enforce a cross-validation to obtain stats, even when no search is necessary (default: 0)
%
% Out:
%   Model : the best model, as determined through cross-validation and and grid search 
%           w.r.t. to a measure of the overall performance of the trainer/tester combination
%
%   Stats : additional statistics over all tested argument combinations, as produced by the metric
%
% See also:
%   utl_crossval(), utl_gridsearch(), utl_nested_crossval()
%
% Example:
%   trials: [NxF] array of training instances
%   targets: [Nx1] array of labels
%   [model,stats] = utl_searchmodel({trials,targets}, 'args',{'svmlinear' search(2.^(-5:2:15))})  % find the best-performing linear SVM classifier on the data
%
%                               Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                               2010-04-22

opts = hlp_varargin2struct(varargin, 'trainer',@ml_train, 'tester',@ml_predict, 'argform','clauses', 'partitioner',@utl_default_partitioner, 'forcestats',0);

if opts.forcestats || is_needing_search(opts.argform,opts.args)
    % do a grid search over cross-validation outcomes, for every specified argument combination
    [stats.bestidx,stats.inputs,stats.outputs] = utl_gridsearch(opts.argform, ...
        @(varargin)utl_crossval(data,opts,'args',varargin),opts.args{:});
else
    % produce dummy stats if neither necessary nor forced
    stats = struct('bestidx',1, 'inputs',{{opts.args}}, 'outputs',{{NaN,struct()}});
end

% using the best args, send the complete data through the partitioner & compute a model on it
bestargs = stats.inputs{stats.bestidx};
data = opts.partitioner(data,1:opts.partitioner(data,[],bestargs),bestargs);
model = opts.trainer(data,bestargs{:});