function icSpectralDecomposition = make_spectral_decomposition(EEG, comps, numberOfSpctralTemplates, freqs, numberOfFrequencies, downsamplingRate)
% icSpectralDecomposition = make_spectral_decomposition(EEG, comps, numberOfSpctralTemplates, freqs, numberOfFrequencies, downsamplingRate)
% freqs can be either a range, like [3 100] (which is default if not given)
% or if there are nore than two elements in freqs, they will be used a individual frequencies.
% when a range is given 'numberOfFrequencies' log spaces frequencies are
% used for the decomposition.
%
% by default 10 templates are produced (or can be set by numberOfSpctralTemplates)
%
% comps contains the component numbers. Each component is processed separately (spectral
% decomposition are only within each IC). For each of this IC decompositions a separate sturcture
% is placed in icSpectralDecomposition array.
%
% output:
%
% icSpectralDecomposition: is a structure array with fields containing
% decomposiion information. Among these some are more important:
% mix and unmix fields which contain spectral ICA matrices.
%
% activity field contains the activation of each template. It contains some
% Infs or Nans for the beginning and end of the input data where wavelets
% could not be applied or some power was near zero...
% the 'badFrames' field contains IDs for all these nan or inf values. If
% you want a clean activation (for example for correlation or third-level
% ICA). You can remove these unwanted values:
%
%   cleanActivity = icSpectralDecomposition(1).activation;
%   cleanActivity(:,icSpectralDecomposition(1).badFrames) = [];
%


if nargin<3
    numberOfSpctralTemplates = 10;
end;

if nargin<4
    freqs = [3 100]; % default
end;

if nargin<5
    numberOfFrequencies = 25;
end;

if nargin<6
    downsamplingRate = 1;
end;

if length(freqs) == 2
    frequencies = logspace(log10(freqs(1)), log10(freqs(2)), numberOfFrequencies);
else % if there are several values in freqs, we assume that they are desired freqquencies
    frequencies = freqs;
end;

% increase the cycles for higher frequencies, for smoother results, go from
% 3 to 10 cycles
% cycles = logspace(log10(3),log10(10),length(frequencies));
%
% for i=1:length(frequencies)
%     [wavelet(i),cyclesDummy,freqresol,timeresol] = dftfilt3( frequencies(i), 3, EEG.srate);
% end;

[wavelet,cycles,freqresol,timeresol] = dftfilt3( frequencies, [3 20], EEG.srate);

for compId = 1:length(comps)
    
    comp = comps(compId);
    powerForWavelet = power_for_wavelet(EEG, comp, wavelet);
    
    meanPower = mean(powerForWavelet,2);
    
    logPowerModulation = log(powerForWavelet ./ repmat(meanPower, 1, size(powerForWavelet, 2))) / log(10) ;
    
    
    badIds = find(isnan(logPowerModulation) |isinf(logPowerModulation)  | logPowerModulation ==0);
    [i badFrames] = ind2sub( size(logPowerModulation), badIds);
    cleanedModulation = logPowerModulation;
    cleanedModulation(:,badFrames) = [];
    
    % down-sample only for calculating ICA
    cleanedModulation = cleanedModulation(:, 1:downsamplingRate:end);
    
    % perform PCA if necessary (i.e. number of spectral templates requested is less).
    if numberOfSpctralTemplates < size(cleanedModulation,1)
        [wts sph] = binica(cleanedModulation, 'extended', numberOfSpctralTemplates, 'stop' ,1.0e-7,'maxsteps',1000, 'pca', numberOfSpctralTemplates);
    else
        [wts sph] = binica(cleanedModulation, 'extended', numberOfSpctralTemplates, 'stop' ,1.0e-7,'maxsteps',1000);
    end;
    
    
    icSpectralDecomposition(compId).meanPower = meanPower;
    
    icSpectralDecomposition(compId).activation = wts * sph * logPowerModulation;
    
    % clean means withouth Infs or Nans
    %icSpectralDecomposition.cleanActivity = wts * sph * cleanedModulation;
    
    
    icSpectralDecomposition(compId).wavelet = wavelet;
    icSpectralDecomposition(compId).frequencies = frequencies;
    icSpectralDecomposition(compId).cycles = cycles;
    icSpectralDecomposition(compId).unmix = wts * sph;
    icSpectralDecomposition(compId).mix = pinv(wts*sph);
    
    % for scalpmap
    icSpectralDecomposition(compId).chanlocs = EEG.chanlocs;
    icSpectralDecomposition(compId).scalpMap = EEG.icawinv(:, comp);
    icSpectralDecomposition(compId).comp = comp;
    icSpectralDecomposition(compId).channelIcaWeights = EEG.icaweights;
    icSpectralDecomposition(compId).channelIcaSphere = EEG.icasphere;
    
    
    icSpectralDecomposition(compId).numberOfFrequencies = numberOfFrequencies;
    icSpectralDecomposition(compId).numberOfSpctralTemplates = numberOfSpctralTemplates;
    icSpectralDecomposition(compId).badFrames = badFrames;
    
    % for extra subject, dataset, group... info
    icSpectralDecomposition(compId).subject = EEG.subject;
    icSpectralDecomposition(compId).group = EEG.subject;
    icSpectralDecomposition(compId).condition = EEG.condition;
    icSpectralDecomposition(compId).setname = EEG.setname;
end;