function result = para_csp(data, mode, varargin)
% Result = para_csp(Input-Data, Operation-Mode, Options...)
% Paradigm for oscillatory processes using the Common Spatial Patterns (CSP) algorithm.
%
% The CSP paradigm is based on the design of the Berlin Brain-Computer Interface (BBCI) [1], more comprehensively described in [2], 
% which is mainly controlled by (sensori-)motor imagery. The features exploited by this paradigm in its original form are Event-Related Synchronization 
% and Desynchronization [3] localized in the (sensori-)motor cortex, but the paradigm is not restricted to these applications. CSP was originally 
% introduced in [5] and first applied to EEG in [6].
%
% Due to its simplicity, speed and relative robustness, CSP is the bread-and-butter paradigm for oscillatory processes, and if nothing else, can be used
% to get a quick estimate of whether the data contains information of interest or not. Like para_bandpower, CSP uses log-variance features over a single 
% non-adapted frequency range (which may have multiple peaks), and neither temporal structure (variations) in the signal is captured, nor are interactions 
% between frequency bands. The major strength of the paradigm is its adaptive spatial filter, which is computed using the CSP algorithm.
%
% The paradigm is implemented as a standard sequence of signal (pre-)processing (spatial/spectral filtering), feature extraction, and machine learning.
% The first preprocessing step is frequency filtering, followed by an adaptively learned spatial filter (which is the defining propery of the paradigm),
% followed by log-variance feature extraction and finally a (usually simple) machine learning step applied to the log-variance features. The spatial filtering
% projects the channels of the original signal down to a small set of (usually 4-6) surrogate channels, where the (linear) mapping is optimized such that 
% the variance in these channels is maximally informative w.r.t. to the prediction task. The CSP filters can be obtained from the per-class signal covariance 
% matrices by solving a generalized eigenvalue problem (of the form [V,D]=eig(Cov1,Cov1+Cov2)). CSP can also be applied to independent components to rate their
% importance or for better artifact robustness. A wide range of classifiers can be used with CSP features, the most commonly used one being LDA.
% There exists a large corpus of CSP variants and extensions, mostly to give better control over spectral filtering, including multiband CSP (para_multiband_csp),
% Spectrally Weighted CSP (para_speccsp), Invariant CSP, Common Spatio-Spectral Patterns (CSSP), Common Sparse Spectral Spatial Pattern (CSSSP), Regularized CSP,
% and several others. A more advanced (but also computationally more costly) paradigm than CSP is the Dual-Augmented Lagrange Paradigm (para_dal/para_dal_hf).
% The length of the data epoch and the choice of a frequency band (defaulting to motor imagery time scales and frequency ranges) are the parameters that are most
% commonly tuned to the task, both of which can also be found via a small parameter search.
% 
% Some application areas include detection of major brain rhythm modulations (e.g. alpha, beta), for example related to relaxation/stress, 
% aspects of workload, sensori-motor imagery, visual processing vs. idling and other idle-rhythm-related questions, or emotion recognition. See also [4].
%
% Example: Consider a calibration dataset in which a subject is in varying degrees of relaxedness throughout different time periods. Markers 'relaxed',
% 'non-relaxed' and 'stressed' have been inserted in 2 second intervals into these data ranges. The goal is to learn a model which can predict these
% states, using alpha rhythm features.
%  
%   calib = io_loadset('datasets/demos/relaxation.eeg')
%   [loss,model,stats] = bci_train({'data',calib, 'paradigm',@para_csp}, 'events',{'relaxed','non-relaxed','stressed'}, 'spectrum',[7 15], 'epoch',[-2 2])
%
%
% In:
%   Input-Set       : some data set
%
%   Operation-Mode  : 'preprocessing', to pre-process the Input-Set according to the paradigm (and parameters)
%                     'learning', to learn a predictive model from a pre-processed Input-Set (and parameters)
%                     'prediction', to make predictions given a pre-processed Input-Set and a Model
%
%   User Options... : optional parameters of the paradigm; name-value pairs 
%                     --- for preprocessing (see flt_default)
%                     * 'spectrum':  spectral filter (default: [7 30])
%                     * 'srate':   sampling rate, in Hz (default: 200)
%                     * 'epoch':  epoch window, in seconds relative to event (default: [0.5 3.5])
%                     * 'events':  event types to determine epochs and the target variable (user-specified)
%                     * others, by default turned off: see flt_default
%                     --- for learning
%                     * 'patterns': the number of pattern pairs that CSP should produce; feature-space dimension will be twice this number (default: 3)
%                     * 'learner': learning function to be used on the features; (default: 'lda')
%                     --- for prediction
%                     * 'model': the detector model (as produced during the learning)
% Out:
%   Result : depends on the Operation-Mode; 
%            * for 'preprocessing', this is the preprocessed dataset
%            * for 'learning', this is the learned model
%            * for 'prediction', this is the predictions produced by the model, given the data
%
% References:
%  [1] Blankertz, B., Dornhege, G., Krauledat, M., Mller, K., and Curio,G. 
%      "The non-invasive Berlin Brain-Computer interface: Fast acquisition of effective performance in untrained subjects."
%      NeuroImage 37, 2 (Aug. 2007), 539?550.
%  [2] Benjamin Blankertz, Ryota Tomioka, Steven Lemm, Motoaki Kawanabe, and Klaus-Robert Mueller. 
%      "Optimizing spatial filters for robust EEG single-trial analysis." 
%      IEEE Signal Process Mag, 25(1):41-56, January 2008
%  [3] Pfurtscheller, G., and da Silva, L. "Event-related EEG/MEG synchronizaion and desynchronization: basic principles." 
%      Clin Neurophysiol 110 (1999), 1842-1857.
%  [4] Buzsaki, G., "Rhythms of the brain"
%      Oxford University Press US, 2006
%  [5] Fukunaga K., "Introduction to Statistical Pattern Recognition" 
%      Academic Press, Computer Science and Scientific Computing Series, 1990 
%  [6] Ramoser, H., Gerking, M., and Pfurtscheller, G. "Optimal spatial filtering of single trial EEG during imagined hand movement." 
%      IEEE Trans. Rehab. Eng 8 (2000), 446, 441.
%
%                           Christian Kothe, Swartz Center for Computational Neuroscience, UCSD
%                           2010-04-29


result = para_skeleton(data, mode, hlp_varargin2struct(varargin, ...
	... % preprocessing defaults:
    'spectrum',[7 30], 'epoch',[0.5 3.5], 'srate',200, ...
    ... % learning defaults:
    'patterns',3, 'learner','lda', ...
    'featureadapt',@csp_train,'featureextract',@csp_predict),'binary','featureadapt');


% obtain CSP spatial filters
function model = csp_train(data,opts)
for k=1:2
    trials{k} = exp_eval(set_picktrials(data,'rank',k),1);
    covar{k} = cov(reshape(trials{k}.data,size(trials{k}.data,1),[])');
    covar{k}(~isfinite(covar{k})) = 0;
end
[V,D] = eig(covar{1},covar{1}+covar{2}); %#ok<NASGU>
model.filters = V(:,[1:opts.patterns end-opts.patterns+1:end]);
P = inv(V);
model.patterns = P([1:opts.patterns end-opts.patterns+1:end],:);

% extract filtered log-variance features
function X = csp_predict(data,model)
X = zeros(size(data.data,3),size(model.filters,2));
for t=1:size(data.data,3)
    X(t,:) = log(var(data.data(:,:,t)'*model.filters)); end
